#!/usr/bin/env python

from importlib import import_module

from django.apps import apps
from django.db import connection
from django.test import TestCase
from django.db.migrations.executor import MigrationExecutor

from pulp_ansible.app.models import AnsibleDistribution, AnsibleRepository


class TestMigrations(TestCase):

    @property
    def app(self):
        return 'ansible'

    migrate_from = None
    migrate_to = None

    def setUp(self):
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)
        self.migrate_from = [(self.app, self.migrate_from)]
        self.migrate_to = [(self.app, self.migrate_to)]
        executor = MigrationExecutor(connection)
        old_apps = executor.loader.project_state(self.migrate_from).apps

        # Reverse to the original migration
        executor.migrate(self.migrate_from)

        self.setUpBeforeMigration(old_apps)

        # Run the migration to test
        executor = MigrationExecutor(connection)
        executor.loader.build_graph()  # reload.
        executor.migrate(self.migrate_to)

        self.apps = executor.loader.project_state(self.migrate_to).apps

    def setUpBeforeMigration(self, apps):
        pass


class TestCrossRepoSearchMigration(TestMigrations):

    migrate_from = '0050_crossrepositorycollectionversionindex'
    migrate_to = '0051_cvindex_build'

    def setUpBeforeMigration(self, apps):
        # make a repository
        repo = AnsibleRepository.objects.create(name='foobar')

        # make a distro
        distro = AnsibleDistribution.objects.create(name='foobar', base_path='foobar', repository=repo)

    def test_nothing(self):
        pass



# Generated by Django 3.2.19 on 2023-06-01 18:41

from django.db import migrations, models
from semantic_version import Version


def parse_versions(apps, *args, **kwargs):
    CollectionVersion = apps.get_model('ansible', 'CollectionVersion')

    to_update = []
    fields = ["version_major", "version_minor", "version_patch", "version_prerelease"]
    for cv in CollectionVersion.objects.all().only("version"):
        v = Version(cv.version)

        cv.version_major = v.major
        cv.version_minor = v.minor
        cv.version_patch = v.patch
        cv.version_prerelease = ".".join(v.prerelease)

        to_update.append(cv)

        if len(to_update) > 1000:
            CollectionVersion.objects.bulk_update(to_update, fields)
            to_update = []

    CollectionVersion.objects.bulk_update(to_update, fields)


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0053_collectiondownloadcount'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionversion',
            name='version_major',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='version_minor',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='version_patch',
            field=models.IntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='version_prerelease',
            field=models.CharField(default='', max_length=128),
            preserve_default=False,
        ),
        migrations.RunPython(
            code=parse_versions,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.AlterField(
            model_name="collectiondownloadcount",
            name="name",
            field=models.CharField(db_index=True, editable=False, max_length=64),
        ),
        migrations.AlterField(
            model_name="collectiondownloadcount",
            name="namespace",
            field=models.CharField(db_index=True, editable=False, max_length=64),
        ),
    ]



def test_client_config_distro(
    ansible_repo, ansible_distribution_factory, ansible_client_configuration_api_client
):
    distro = ansible_distribution_factory(ansible_repo)
    res = ansible_client_configuration_api_client.read(path=distro.base_path)
    assert res.default_distribution_path == distro.base_path


def test_client_config_default_distro(
    ansible_repo, ansible_distribution_factory, ansible_client_default_configuration_api_client
):
    res = ansible_client_default_configuration_api_client.read()
    assert res.default_distribution_path is None



# Generated by Django 2.2.13 on 2020-08-11 17:51

import django
from django.db import migrations, models, transaction


def update_pulp_type(apps, schema_editor):
    with transaction.atomic():
        Remote = apps.get_model('core', 'Remote')
        Remote.objects.filter(pulp_type="ansible.ansible").update(pulp_type="ansible.role")


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0039_change_download_concurrency'),
        ('ansible', '0020_auto_20200810_1926'),
    ]

    operations = [
        migrations.RenameModel('AnsibleRemote', 'RoleRemote'),
        migrations.AlterField(
            model_name='roleremote',
            name='remote_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_roleremote', serialize=False, to='core.Remote'),
        ),
        migrations.RunPython(update_pulp_type),
    ]



import uuid
import pytest
import numpy as np
from PIL import Image
import time

from orionutils.generator import build_collection, randstr

from pulpcore.client.pulp_ansible import (
    AnsibleCollectionsApi,
    AnsibleRepositorySyncURL,
    ApiClient,
    ContentCollectionDeprecationsApi,
    ContentCollectionMarksApi,
    ContentCollectionSignaturesApi,
    ContentCollectionVersionsApi,
    ContentNamespacesApi,
    ContentRolesApi,
    DistributionsAnsibleApi,
    PulpAnsibleApiV3CollectionsVersionsApi,
    RepositoriesAnsibleApi,
    RepositoriesAnsibleVersionsApi,
    RemotesCollectionApi,
    RemotesGitApi,
    RemotesRoleApi,
    PulpAnsibleApiV3CollectionsApi,
    PulpAnsibleApiV3NamespacesApi,
    PulpAnsibleApiV3PluginAnsibleClientConfigurationApi,
    PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexVersionsApi,
    PulpAnsibleDefaultApiV3PluginAnsibleClientConfigurationApi,
    PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexApi,
    PulpAnsibleApiV3PluginAnsibleContentNamespacesApi,
    PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi,
    PulpAnsibleDefaultApiV3PluginAnsibleSearchCollectionVersionsApi,
)

from pulp_ansible.tests.functional.constants import ANSIBLE_FIXTURE_URL


# Bindings API Fixtures


@pytest.fixture(scope="session")
def ansible_bindings_client(_api_client_set, bindings_cfg):
    """Provides the Ansible bindings client object."""
    api_client = ApiClient(bindings_cfg)
    _api_client_set.add(api_client)
    yield api_client
    _api_client_set.remove(api_client)


@pytest.fixture(scope="session")
def ansible_collections_api_client(ansible_bindings_client):
    """Provides the Ansible Collections API client object."""
    return AnsibleCollectionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_client_configuration_api_client(ansible_bindings_client):
    """Provides the Ansible Collections API client object."""
    return PulpAnsibleApiV3PluginAnsibleClientConfigurationApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_client_default_configuration_api_client(ansible_bindings_client):
    """Provides the Ansible Collections API client object."""
    return PulpAnsibleDefaultApiV3PluginAnsibleClientConfigurationApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_collection_signatures_client(ansible_bindings_client):
    """Provides the Ansible Collection Signatures API client object."""
    return ContentCollectionSignaturesApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_collection_mark_client(ansible_bindings_client):
    """Provides the Ansible Collection Marks API client object."""
    return ContentCollectionMarksApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_repo_api_client(ansible_bindings_client):
    """Provides the Ansible Repository API client object."""
    return RepositoriesAnsibleApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_repo_version_api_client(ansible_bindings_client):
    """Provides the Ansible Repository Version API client object."""
    return RepositoriesAnsibleVersionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_distro_api_client(ansible_bindings_client):
    """Provides the Ansible Distribution API client object."""
    return DistributionsAnsibleApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_collection_deprecations_api_client(ansible_bindings_client):
    """Provides the Ansible Deprecations API client object."""
    return ContentCollectionDeprecationsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_collection_version_api_client(ansible_bindings_client):
    """Provides the Ansible Content Collection Version API client object."""
    return ContentCollectionVersionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_remote_collection_api_client(ansible_bindings_client):
    """Provides the Ansible Collection Remotes API client object."""
    return RemotesCollectionApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_remote_role_api_client(ansible_bindings_client):
    """Provides the Ansible Role Remotes API client object."""
    return RemotesRoleApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_remote_git_api_client(ansible_bindings_client):
    """Provides the Ansible Git Remotes API client object."""
    return RemotesGitApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_namespaces_api_client(ansible_bindings_client):
    """Provides the Ansible Content Namespaces API client object."""
    return ContentNamespacesApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def ansible_roles_api_client(ansible_bindings_client):
    """Provides the Ansible Content Roles API client object."""
    return ContentRolesApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_namespaces_api_client(ansible_bindings_client):
    """Provides the *deprecated* Galaxy V3 Namespace API client object."""
    return PulpAnsibleApiV3NamespacesApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_content_collection_index_api(ansible_bindings_client):
    """Provides the Galaxy V3 Collections Index API client object."""
    return PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_plugin_namespaces_api_client(ansible_bindings_client):
    """Provides the Galaxy V3 Namespace API client object."""
    return PulpAnsibleApiV3PluginAnsibleContentNamespacesApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_collection_versions_api_client(ansible_bindings_client):
    """Provides the *deprecated* Galaxy V3 Collection Versions API client object."""
    return PulpAnsibleApiV3CollectionsVersionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_collections_api_client(ansible_bindings_client):
    """Provides the *deprecated* Galaxy V3 Collections API client object."""
    return PulpAnsibleApiV3CollectionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_content_collections_index_api_client(ansible_bindings_client):
    """Provides the Galaxy V3 Collection API client object."""
    return PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_content_collections_index_versions_api_client(ansible_bindings_client):
    """Provides the Galaxy V3 Collection Versions API client object."""
    return PulpAnsibleApiV3PluginAnsibleContentCollectionsIndexVersionsApi(ansible_bindings_client)


@pytest.fixture(scope="session")
def galaxy_v3_default_search_api_client(ansible_bindings_client):
    """Provides the Galaxy V3 Search API client object."""
    return PulpAnsibleDefaultApiV3PluginAnsibleSearchCollectionVersionsApi(ansible_bindings_client)


# Object Generation Fixtures


@pytest.fixture
def ansible_repo(ansible_repo_factory):
    """Creates an Ansible Repository and deletes it at test cleanup time."""
    return ansible_repo_factory()


@pytest.fixture(scope="class")
def ansible_repo_factory(ansible_repo_api_client, gen_object_with_cleanup):
    """A factory that creates an Ansible Repository and deletes it at test cleanup time."""

    def _ansible_repo_factory(**kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        return gen_object_with_cleanup(ansible_repo_api_client, kwargs)

    return _ansible_repo_factory


@pytest.fixture(scope="class")
def ansible_sync_factory(ansible_repo_api_client, ansible_repo_factory, monitor_task):
    """A factory to perform a sync on an Ansible Repository and return its updated data."""

    def _sync(ansible_repo=None, **kwargs):
        body = AnsibleRepositorySyncURL(**kwargs)
        if ansible_repo is None:
            ansible_repo = ansible_repo_factory()
        monitor_task(ansible_repo_api_client.sync(ansible_repo.pulp_href, body).task)
        return ansible_repo_api_client.read(ansible_repo.pulp_href)

    return _sync


@pytest.fixture(scope="class")
def ansible_distribution_factory(ansible_distro_api_client, gen_object_with_cleanup):
    """A factory to generate an Ansible Distribution with auto-cleanup."""

    def _ansible_distribution_factory(repository=None, **kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        kwargs.setdefault("base_path", str(uuid.uuid4()))
        if repository:
            kwargs.setdefault("repository", repository.pulp_href)
        return gen_object_with_cleanup(ansible_distro_api_client, kwargs)

    return _ansible_distribution_factory


@pytest.fixture(scope="class")
def ansible_role_remote_factory(
    bindings_cfg, ansible_remote_role_api_client, gen_object_with_cleanup
):
    """A factory to generate an Ansible Collection Remote with auto-cleanup."""

    def _ansible_role_remote_factory(**kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        kwargs.setdefault("url", ANSIBLE_FIXTURE_URL)
        if kwargs.pop("include_pulp_auth", False):
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        return gen_object_with_cleanup(ansible_remote_role_api_client, kwargs)

    return _ansible_role_remote_factory


@pytest.fixture(scope="class")
def ansible_collection_remote_factory(
    bindings_cfg, ansible_remote_collection_api_client, gen_object_with_cleanup
):
    """A factory to generate an Ansible Collection Remote with auto-cleanup."""

    def _ansible_collection_remote_factory(**kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        if kwargs.pop("include_pulp_auth", False):
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        return gen_object_with_cleanup(ansible_remote_collection_api_client, kwargs)

    return _ansible_collection_remote_factory


@pytest.fixture(scope="class")
def ansible_git_remote_factory(
    bindings_cfg, ansible_remote_git_api_client, gen_object_with_cleanup
):
    """A factory to generate an Ansible Git Remote with auto-cleanup."""

    def _ansible_git_remote_factory(**kwargs):
        kwargs.setdefault("name", str(uuid.uuid4()))
        if kwargs.pop("include_pulp_auth", False):
            kwargs["username"] = bindings_cfg.username
            kwargs["password"] = bindings_cfg.password
        return gen_object_with_cleanup(ansible_remote_git_api_client, kwargs)

    return _ansible_git_remote_factory


@pytest.fixture
def build_and_upload_collection(ansible_collection_version_api_client, monitor_task):
    """A factory to locally create, build, and upload a collection."""

    def _build_and_upload_collection(ansible_repo=None, **kwargs):
        collection = build_collection("skeleton", **kwargs)
        body = {"file": collection.filename}
        if ansible_repo:
            body["repository"] = ansible_repo.pulp_href
        response = ansible_collection_version_api_client.create(**body)
        task = monitor_task(response.task)
        collection_href = [
            href for href in task.created_resources if "content/ansible/collection_versions" in href
        ]
        return collection, collection_href[0]

    return _build_and_upload_collection


@pytest.fixture
def ansible_dir_factory(tmp_path, bindings_cfg):
    """A factory to create a local ansible.cfg file with authentication"""

    def _ansible_dir_factory(server):
        username = bindings_cfg.username
        password = bindings_cfg.password

        ansible_cfg = (
            "[galaxy]\nserver_list = pulp_ansible\n\n"
            "[galaxy_server.pulp_ansible]\n"
            f"url = {server}\n"
        )

        if username is not None and password is not None:
            ansible_cfg += f"username = {username}\npassword = {password}\n"

        cfg_file = tmp_path / "ansible.cfg"
        cfg_file.write_text(ansible_cfg, encoding="UTF_8")
        return tmp_path

    return _ansible_dir_factory


@pytest.fixture
def random_image_factory(tmp_path):
    """Factory to produce a random 100x100 image."""

    def _random_image():
        imarray = np.random.rand(100, 100, 3) * 255
        im = Image.fromarray(imarray.astype("uint8")).convert("RGBA")
        path = tmp_path / f"{randstr()}.png"
        im.save(path)
        return path

    return _random_image


# Utility fixtures


@pytest.fixture(scope="session")
def wait_tasks(tasks_api_client):
    """Polls the Task API until all tasks for a resource are in a completed state."""

    def _wait_tasks(resource):
        pending_tasks = tasks_api_client.list(
            state__in=["running", "waiting"], reserved_resources=resource
        )
        while pending_tasks.count:
            time.sleep(1)
            pending_tasks = tasks_api_client.list(
                state__in=["running", "waiting"], reserved_resources=resource
            )

    return _wait_tasks



import unittest
import requests

from pulpcore.client.pulp_ansible import (
    PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi,
    PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexVersionsApi,
)
from pulp_smash import config

from pulp_ansible.tests.functional.utils import (
    gen_ansible_remote,
    is_galaxy_ng_installed,
)
from pulp_ansible.tests.functional.utils import SyncHelpersMixin, TestCaseUsingBindings
from pulp_ansible.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulpcore import ContentguardsContentRedirectApi, ApiClient as CoreApiClient


class CollectionDownloadTestCase(TestCaseUsingBindings, SyncHelpersMixin):
    """Test content redirect guard."""

    def setUp(self):
        """Set up the content guard tests."""
        self.requirements_file = "collections:\n  - testing.k8s_demo_collection"
        body = gen_ansible_remote(
            url="https://galaxy.ansible.com",
            requirements_file=self.requirements_file,
            sync_dependencies=False,
        )
        self.remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, self.remote.pulp_href)

        self.first_repo = self._create_repo_and_sync_with_remote(self.remote)
        self.distribution = self._create_distribution_from_repo(self.first_repo)

    def _get_download_url(self, namespace, name):
        collection_api = PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi(self.client)
        versions_api = PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexVersionsApi(
            self.client
        )

        collection = collection_api.read(
            self.distribution.base_path,
            name,
            namespace,
        )

        version = versions_api.read(
            self.distribution.base_path, name, namespace, collection.highest_version["version"]
        )

        return version.download_url

    @unittest.skipIf(
        is_galaxy_ng_installed(),
        (
            "This test fails with galaxy_ng installed because galaxy_ng ensures that "
            "a contentguard is always present on ansible repos."
        ),
    )
    def test_download(self):
        """Test that downloads without content guards work correctly."""
        # Using requests for these tests because the client follows redirects and I need to
        # inspect the URL that gets redirected to
        download_url = self._get_download_url("testing", "k8s_demo_collection")
        cfg = config.get_config()
        response = requests.get(download_url, auth=tuple(cfg.pulp_auth), allow_redirects=False)

        # verify that the download url redirects to the content app
        assert response.is_redirect
        content_app_url = response.headers["Location"]

        # verify that the collection can be downloaded without authentication
        assert "validate_token" not in content_app_url

        collection = requests.get(content_app_url)
        assert collection.status_code == 200

    def test_download_with_content_guard(self):
        """Test that downloads with content guards work correctly."""
        # Using requests for these tests because the client follows redirects and I need to
        # inspect the URL that gets redirected to
        # setup content guard
        cfg = config.get_config()
        configuration = cfg.get_bindings_config()
        core_client = CoreApiClient(configuration)
        guard_client = ContentguardsContentRedirectApi(core_client)
        guard = guard_client.create({"name": "test-content-guard"})

        self.addCleanup(guard_client.delete, guard.pulp_href)

        self.distributions_api.partial_update(
            self.distribution.pulp_href, {"content_guard": guard.pulp_href}
        )

        download_url = self._get_download_url("testing", "k8s_demo_collection")
        response = requests.get(download_url, auth=tuple(cfg.pulp_auth), allow_redirects=False)

        # verify that the download url redirects to the content app
        assert response.is_redirect
        content_app_url = response.headers["Location"]

        # verify that token is present
        assert "validate_token" in content_app_url

        collection = requests.get(content_app_url)
        assert collection.status_code == 200

        # make an unauthenticated call to the content app and verify that it gets
        # rejected
        collection = requests.get(content_app_url.split("?")[0])
        assert collection.status_code == 403



"""Tests that Collections can be uploaded to  Pulp with the ansible-galaxy CLI."""

import pytest
import random
import string
import subprocess


@pytest.mark.parallel
def test_upload_collection(
    ansible_repo_api_client,
    galaxy_v3_collections_api_client,
    galaxy_v3_collection_versions_api_client,
    ansible_repo_factory,
    ansible_distribution_factory,
    ansible_dir_factory,
    wait_tasks,
):
    """Test whether ansible-galaxy can upload a Collection to Pulp."""
    repository = ansible_repo_factory()
    distribution = ansible_distribution_factory(repository=repository)
    ansible_dir = ansible_dir_factory(distribution.client_url)

    collection_name = "".join(random.choices(string.ascii_lowercase, k=26))
    subprocess.run(
        (
            "ansible-galaxy",
            "collection",
            "init",
            "--init-path",
            ansible_dir,
            f"pulp.{collection_name}",
        ),
        cwd=ansible_dir,
        check=True,
    )

    collection_meta = ansible_dir / "pulp" / collection_name / "meta"
    collection_meta.mkdir()
    runtime_yml = collection_meta / "runtime.yml"
    runtime_yml.write_text('requires_ansible: ">=2.9"\n')

    subprocess.run(
        (
            "ansible-galaxy",
            "collection",
            "build",
            "--output-path",
            ansible_dir,
            ansible_dir / "pulp" / collection_name,
        ),
        cwd=ansible_dir,
        check=True,
    )

    assert repository.latest_version_href.endswith("/0/")
    collections = galaxy_v3_collections_api_client.list(distribution.base_path)
    assert collections.meta.count == 0

    subprocess.run(
        (
            "ansible-galaxy",
            "collection",
            "publish",
            "-c",
            ansible_dir / f"pulp-{collection_name}-1.0.0.tar.gz",
        ),
        cwd=ansible_dir,
        check=True,
    )
    wait_tasks(repository.pulp_href)

    repository = ansible_repo_api_client.read(repository.pulp_href)
    assert repository.latest_version_href.endswith("/1/")
    collections = galaxy_v3_collections_api_client.list(distribution.base_path)
    assert collections.meta.count == 1  # We uploaded 1 collection
    version = galaxy_v3_collection_versions_api_client.read(
        collection_name, "pulp", distribution.base_path, "1.0.0"
    )
    assert version.requires_ansible == ">=2.9"



import asyncio
import hashlib
import json
import logging
import tarfile
from collections import defaultdict
from gettext import gettext as _
from operator import attrgetter
from urllib.parse import urljoin
from uuid import uuid4

import yaml
from aiohttp.client_exceptions import ClientError, ClientResponseError
from asgiref.sync import sync_to_async
from async_lru import alru_cache
from django.conf import settings
from django.db import transaction
from django.db.models import F, Q
from django.db.utils import IntegrityError
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from galaxy_importer.collection import CollectionFilename
from galaxy_importer.collection import import_collection as process_collection
from galaxy_importer.collection import sync_collection
from galaxy_importer.exceptions import ImporterError
from git import GitCommandError, Repo
from pulpcore.plugin.exceptions import DigestValidationError
from pulpcore.plugin.models import (
    Artifact,
    ContentArtifact,
    CreatedResource,
    ProgressReport,
    PulpTemporaryFile,
    Remote,
    RepositoryContent,
    RepositoryVersion,
    Task,
)
from pulpcore.plugin.stages import (
    ArtifactDownloader,
    ArtifactSaver,
    ContentSaver,
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    QueryExistingArtifacts,
    QueryExistingContents,
    RemoteArtifactSaver,
    ResolveContentFutures,
    Stage,
)
from rest_framework.serializers import ValidationError
from semantic_version import SimpleSpec, Version
from semantic_version.base import Always

from pulp_ansible.app.constants import PAGE_SIZE
from pulp_ansible.app.models import (
    AnsibleCollectionDeprecated,
    AnsibleNamespace,
    AnsibleNamespaceMetadata,
    AnsibleRepository,
    Collection,
    CollectionImport,
    CollectionRemote,
    CollectionVersion,
    CollectionVersionMark,
    CollectionVersionSignature,
    Tag,
)
from pulp_ansible.app.serializers import CollectionVersionSerializer
from pulp_ansible.app.tasks.utils import (
    RequirementsFileEntry,
    get_file_obj_from_tarball,
    parse_collections_requirements_file,
    parse_metadata,
)

log = logging.getLogger(__name__)


# semantic_version.SimpleSpec interpretes "*" as ">=0.0.0"
class AnsibleSpec(SimpleSpec):
    def __init__(self, expression):
        super().__init__(expression)
        if self.expression == "*":
            self.clause = Always()


async def declarative_content_from_git_repo(remote, url, git_ref=None, metadata_only=False):
    """Returns a DeclarativeContent for the Collection in a Git repository."""
    if git_ref:
        try:
            gitrepo = Repo.clone_from(
                url, uuid4(), depth=1, branch=git_ref, multi_options=["--recurse-submodules"]
            )
        except GitCommandError:
            gitrepo = Repo.clone_from(url, uuid4(), multi_options=["--recurse-submodules"])
            gitrepo.git.checkout(git_ref)
    else:
        gitrepo = Repo.clone_from(url, uuid4(), depth=1, multi_options=["--recurse-submodules"])
    commit_sha = gitrepo.head.commit.hexsha
    metadata, artifact_path = sync_collection(gitrepo.working_dir, ".")
    if not metadata_only:
        artifact = Artifact.init_and_validate(artifact_path)
        try:
            await sync_to_async(artifact.save)()
        except IntegrityError:
            artifact = Artifact.objects.get(sha256=artifact.sha256)
        metadata["artifact_url"] = reverse("artifacts-detail", args=[artifact.pk])
        metadata["artifact"] = artifact
    else:
        metadata["artifact"] = None
        metadata["artifact_url"] = None
    metadata["remote_artifact_url"] = "{}/commit/{}".format(url.rstrip("/"), commit_sha)

    artifact = metadata["artifact"]
    try:
        collection_version = await sync_to_async(create_collection_from_importer)(
            metadata, metadata_only=metadata_only
        )
        await sync_to_async(ContentArtifact.objects.get_or_create)(
            artifact=artifact,
            content=collection_version,
            relative_path=collection_version.relative_path,
        )
    except ValidationError as e:
        if e.args[0]["non_field_errors"][0].code == "unique":
            namespace = metadata["metadata"]["namespace"]
            name = metadata["metadata"]["name"]
            version = metadata["metadata"]["version"]
        else:
            raise e
        collection_version = await sync_to_async(CollectionVersion.objects.get)(
            namespace=namespace, name=name, version=version
        )
    if artifact is None:
        artifact = Artifact()
    d_artifact = DeclarativeArtifact(
        artifact=artifact,
        url=metadata["remote_artifact_url"],
        relative_path=collection_version.relative_path,
        remote=remote,
        deferred_download=metadata_only,
    )

    # TODO: docs blob support??

    d_content = DeclarativeContent(
        content=collection_version,
        d_artifacts=[d_artifact],
    )
    return d_content


def sync(remote_pk, repository_pk, mirror, optimize):
    """
    Sync Collections with ``remote_pk``, and save a new RepositoryVersion for ``repository_pk``.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.
        optimize (boolean): Whether to optimize sync or not.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = CollectionRemote.objects.get(pk=remote_pk)
    repository = AnsibleRepository.objects.get(pk=repository_pk)

    is_repo_remote = False
    if repository.remote:
        is_repo_remote = remote.pk == repository.remote.pk

    if not remote.url:
        raise ValueError(_("A CollectionRemote must have a 'url' specified to synchronize."))

    deprecation_before_sync = set()
    for namespace, name in AnsibleCollectionDeprecated.objects.filter(
        pk__in=repository.latest_version().content
    ).values_list("namespace", "name"):
        deprecation_before_sync.add(f"{namespace}.{name}")
    first_stage = CollectionSyncFirstStage(
        remote, repository, is_repo_remote, deprecation_before_sync, optimize
    )
    if first_stage.should_sync is False:
        log.debug(_("no-op: remote wasn't updated since last sync."))
        return

    d_version = AnsibleDeclarativeVersion(first_stage, repository, mirror=mirror)
    repo_version = d_version.create()

    if not repo_version:
        return

    repository.last_synced_metadata_time = first_stage.last_synced_metadata_time
    repository.save(update_fields=["last_synced_metadata_time"])

    to_undeprecate = Q()
    undeprecated = deprecation_before_sync.difference(first_stage.deprecation_after_sync)
    if undeprecated:
        for collection in undeprecated:
            namespace, name = collection.split(".")
            to_undeprecate |= Q(namespace=namespace, name=name)
        deprecated = AnsibleCollectionDeprecated.objects.filter(to_undeprecate)
        RepositoryContent.objects.filter(
            repository=repository, content__in=deprecated
        ).all().update(version_removed=repo_version)


def import_collection(
    temp_file_pk,
    repository_pk=None,
    expected_namespace=None,
    expected_name=None,
    expected_version=None,
):
    """
    Create a Collection from an uploaded artifact and optionally validate its expected metadata.

    This task provides optional validation of the `namespace`, `name`, and `version` metadata
    attributes. If the Artifact fails validation or parsing, the Artifact is deleted and the
    Collection is not created.

    This task performs a CollectionImport object get_or_create() to allow import messages to be
    logged.

    Args:
        temp_file_pk (str): The pk of the PulpTemporaryFile to create the Collection from.

    Keyword Args:
        repository_pk (str): Optional. If specified, a new RepositoryVersion will be created for the
            Repository and any new Collection content associated with it.
        expected_namespace (str): Optional. The namespace is validated against the namespace
            specified in the Collection's metadata. If it does not match a ImporterError is
            raised.
        expected_name (str): Optional. The name is validated against the name specified in the
            Collection's metadata. If it does not match a ImporterError is raised.
        expected_version (str): Optional. The version is validated against the version specified in
            the Collection's metadata. If it does not match a ImporterError is raised.

    Raises:
        ImporterError: If the `expected_namespace`, `expected_name`, or `expected_version` do not
            match the metadata in the tarball.

    """
    CollectionImport.objects.get_or_create(task_id=Task.current().pulp_id)

    temp_file = PulpTemporaryFile.objects.get(pk=temp_file_pk)
    filename = CollectionFilename(expected_namespace, expected_name, expected_version)
    log.info(f"Processing collection from {temp_file.file.name}")
    user_facing_logger = logging.getLogger("pulp_ansible.app.tasks.collection.import_collection")

    try:
        with temp_file.file.open() as artifact_file:
            with tarfile.open(fileobj=artifact_file, mode="r") as tar:
                manifest_data = json.load(
                    get_file_obj_from_tarball(tar, "MANIFEST.json", temp_file.file.name)
                )
                files_data = json.load(
                    get_file_obj_from_tarball(tar, "FILES.json", temp_file.file.name)
                )
            url = _get_backend_storage_url(artifact_file)
            artifact_file.seek(0)
            importer_result = process_collection(
                artifact_file, filename=filename, file_url=url, logger=user_facing_logger
            )
            artifact = Artifact.from_pulp_temporary_file(temp_file)
            temp_file = None
            importer_result["artifact_url"] = reverse("artifacts-detail", args=[artifact.pk])
            collection_version = create_collection_from_importer(importer_result)
            collection_version.manifest = manifest_data
            collection_version.files = files_data
            collection_version.save()

    except ImporterError as exc:
        log.info(f"Collection processing was not successful: {exc}")
        if temp_file:
            temp_file.delete()
        raise
    except Exception as exc:
        user_facing_logger.error(f"Collection processing was not successful: {exc}")
        if temp_file:
            temp_file.delete()
        raise

    ContentArtifact.objects.create(
        artifact=artifact,
        content=collection_version,
        relative_path=collection_version.relative_path,
    )
    CreatedResource.objects.create(content_object=collection_version)

    if repository_pk:
        repository = AnsibleRepository.objects.get(pk=repository_pk)
        content_q = CollectionVersion.objects.filter(pk=collection_version.pk)
        with repository.new_version() as new_version:
            new_version.add_content(content_q)
        CreatedResource.objects.create(content_object=repository)


def create_collection_from_importer(importer_result, metadata_only=False):
    """
    Process results from importer.
    """
    collection_info = importer_result["metadata"]

    with transaction.atomic():
        collection, created = Collection.objects.get_or_create(
            namespace=collection_info["namespace"], name=collection_info["name"]
        )

        tags = collection_info.pop("tags")

        # Remove fields not used by this model
        collection_info.pop("license_file")
        collection_info.pop("readme")

        # the importer returns many None values. We need to let the defaults in the model prevail
        for key in ["description", "documentation", "homepage", "issues", "repository"]:
            if collection_info[key] is None:
                collection_info.pop(key)

        collection_version = CollectionVersion(
            collection=collection,
            **collection_info,
            requires_ansible=importer_result.get("requires_ansible"),
            contents=importer_result["contents"],
            docs_blob=importer_result["docs_blob"],
        )

        serializer_fields = CollectionVersionSerializer.Meta.fields
        data = {k: v for k, v in collection_version.__dict__.items() if k in serializer_fields}
        data["id"] = collection_version.pulp_id
        if not metadata_only:
            data["artifact"] = importer_result["artifact_url"]

        serializer = CollectionVersionSerializer(data=data)

        serializer.is_valid(raise_exception=True)
        collection_version.save()

        for name in tags:
            tag, created = Tag.objects.get_or_create(name=name)
            collection_version.tags.add(tag)

        collection_version.save()  # Save the FK updates
        _update_highest_version(collection_version)
    return collection_version


def rebuild_repository_collection_versions_metadata(
    repository_version_pk, namespace=None, name=None, version=None
):
    """Rebuild metadata for all collection versions in a repo."""
    repov = RepositoryVersion.objects.get(pk=repository_version_pk)

    qs = None
    if namespace or name or version:
        qkwargs = {}
        if namespace:
            qkwargs["namespace"] = namespace
        if name:
            qkwargs["name"] = name
        if version:
            qkwargs["version"] = version
        qs = CollectionVersion.objects.filter(**qkwargs)

    qs = repov.get_content(content_qs=qs)
    with ProgressReport(
        message=_("Rebuild collection version metadata (total)"),
        code="rebuild_metadata.total",
        total=qs.count(),
    ) as ptotal, ProgressReport(
        message=_("Rebuild collection version metadata (failed)"), code="rebuild_metadata.failed"
    ) as pfailed:
        for cv in qs:
            try:
                _rebuild_collection_version_meta(cv)
            except Exception as e:
                pfailed.increment()
                log.exception(e)
            ptotal.increment()


def _rebuild_collection_version_meta(content_object):
    """Rebuild metadata for a single collection version."""

    # Cast to get the CV
    collection_version = content_object.cast()

    # where is the artifact?
    artifact = content_object._artifacts.first()

    # call the importer to re-generate meta
    importer_result = process_collection(
        artifact.file, filename=artifact.file.name, file_url=artifact.file.url, logger=None
    )

    # set the new info and save
    collection_version.requires_ansible = importer_result["requires_ansible"]
    collection_version.docs_blob = importer_result["docs_blob"]
    collection_version.contents = importer_result["contents"]
    collection_version.save()


def _get_backend_storage_url(artifact_file):
    """Get artifact url from pulp backend storage."""
    if (
        settings.DEFAULT_FILE_STORAGE == "pulpcore.app.models.storage.FileSystem"
        or not settings.REDIRECT_TO_OBJECT_STORAGE
    ):
        url = None
    elif settings.DEFAULT_FILE_STORAGE == "storages.backends.s3boto3.S3Boto3Storage":
        parameters = {"ResponseContentDisposition": "attachment%3Bfilename=archive.tar.gz"}
        url = artifact_file.storage.url(artifact_file.name, parameters=parameters)
    elif settings.DEFAULT_FILE_STORAGE == "storages.backends.azure_storage.AzureStorage":
        url = artifact_file.storage.url(artifact_file.name)
    else:
        raise NotImplementedError(
            f"The value settings.DEFAULT_FILE_STORAGE={settings.DEFAULT_FILE_STORAGE} "
            "was not expected"
        )
    return url


def _update_highest_version(collection_version):
    """
    Checks if this version is greater than the most highest one.

    TODO: This function violates content immutability. It must be deprecated.
    """

    qs = collection_version.collection.versions.annotate(prerelease=Q(version_prerelease__ne=""))
    highest_subq = qs.order_by("prerelease", "-version")[0:1].values("pk")
    update_qs = qs.annotate(new_is_highest=Q(pk=highest_subq))
    # To avoid ordering issues, set everything to False first, then update with proper value
    for i in range(2):
        try:
            with transaction.atomic():
                update_qs.update(is_highest=False)
                update_qs.update(is_highest=F("new_is_highest"))
        except IntegrityError:
            log.debug(f"Retrying update_highest_version {collection_version.relative_path}")
            pass
        else:
            return
    log.warning(
        _("Failed to update is_highest for {}, potentially out of date").format(
            collection_version.relative_path
        )
    )


class AnsibleDeclarativeVersion(DeclarativeVersion):
    """
    Subclassed Declarative version creates a custom pipeline for Ansible sync.
    """

    def pipeline_stages(self, new_version):
        """
        Build a list of stages feeding into the ContentUnitAssociation stage.

        This defines the "architecture" of the entire sync.

        Args:
            new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`): The
                new repository version that is going to be built.

        Returns:
            list: List of :class:`~pulpcore.plugin.stages.Stage` instances

        """
        pipeline = [
            self.first_stage,
            QueryExistingArtifacts(),
            ArtifactDownloader(),
            ArtifactSaver(),
            QueryExistingContents(),
            DocsBlobDownloader(),
            AnsibleContentSaver(new_version),
            RemoteArtifactSaver(),
            ResolveContentFutures(),
        ]

        return pipeline


class CollectionSyncFirstStage(Stage):
    """
    The first stage of a pulp_ansible sync pipeline.
    """

    def __init__(self, remote, repository, is_repo_remote, deprecation_before_sync, optimize):
        """
        The first stage of a pulp_ansible sync pipeline.

        Args:
            remote (CollectionRemote): The remote data to be used when syncing
            repository (AnsibleRepository): The repository being syncedself.
            is_repo_remote (bool): True if the remote is the repository's remote.
            deprecation_before_sync (set): Set of deprecations before the sync.
            optimize (boolean): Whether to optimize sync or not.

        """
        super().__init__()
        self.remote = remote
        self.repository = repository
        self.deprecation_before_sync = deprecation_before_sync
        self.deprecation_after_sync = set()
        self.collection_info = parse_collections_requirements_file(remote.requirements_file)
        self.exclude_info = {}
        self.add_dependents = self.collection_info and self.remote.sync_dependencies
        self.signed_only = self.remote.signed_only
        self.already_synced = set()
        self._unpaginated_collection_metadata = None
        self._unpaginated_collection_version_metadata = None
        self.optimize = optimize
        self.last_synced_metadata_time = None
        self.namespaces_seen = set()
        self._unpaginated_namespace_metadata = None

        # Interpret download policy
        self.deferred_download = self.remote.policy != Remote.IMMEDIATE

        # Check if we should sync
        self.should_sync = not is_repo_remote or asyncio.get_event_loop().run_until_complete(
            self._should_we_sync()
        )

    @alru_cache(maxsize=128)
    async def _get_root_api(self, root):
        """
        Returns the root api path and api version.

        Based on https://git.io/JTMxE.
        """
        if root == "https://galaxy.ansible.com" or root == "https://galaxy.ansible.com/":
            root = "https://galaxy.ansible.com/api/"

        downloader = self.remote.get_downloader(url=root)

        try:
            api_data = parse_metadata(await downloader.run())
        except (json.decoder.JSONDecodeError, ClientResponseError):
            if root.endswith("/api/"):
                raise

            root = urljoin(root, "api/")
            downloader = self.remote.get_downloader(url=root)
            api_data = parse_metadata(await downloader.run())

        if "available_versions" not in api_data:
            raise RuntimeError(_("Could not find 'available_versions' at {}").format(root))

        if "v3" in api_data.get("available_versions", {}):
            api_version = 3
        elif "v2" in api_data.get("available_versions", {}):
            api_version = 2
        else:
            raise RuntimeError(_("Unsupported API versions at {}").format(root))

        endpoint = f"{root}v{api_version}"

        return endpoint, api_version

    @alru_cache(maxsize=128)
    async def _get_paginated_collection_api(self, root):
        """
        Returns the collection api path and api version.

        Based on https://git.io/JTMxE.
        """
        endpoint, api_version = await self._get_root_api(root)
        return f"{endpoint}/collections/", api_version

    async def _fetch_collection_version_metadata(self, api_version, collection_version_url):
        downloader = self.remote.get_downloader(url=collection_version_url)
        metadata = parse_metadata(await downloader.run())
        await self._add_collection_version(api_version, collection_version_url, metadata)

    async def _add_collection_version(self, api_version, collection_version_url, metadata):
        """Add CollectionVersion to the sync pipeline."""
        url = metadata["download_url"]
        collection_version = CollectionVersion(
            namespace=metadata["namespace"]["name"],
            name=metadata["collection"]["name"],
            version=metadata["version"],
        )
        cv_unique = attrgetter("namespace", "name", "version")(collection_version)
        fullname, version = f"{cv_unique[0]}.{cv_unique[1]}", cv_unique[2]
        if fullname in self.exclude_info and Version(version) in self.exclude_info[fullname]:
            return
        if cv_unique in self.already_synced:
            return
        self.already_synced.add(cv_unique)

        info = metadata["metadata"]
        signatures = metadata.get("signatures", [])
        marks = metadata.get("marks", [])

        if self.signed_only and not signatures:
            return

        if self.add_dependents:
            dependencies = info["dependencies"]
            tasks = []
            loop = asyncio.get_event_loop()
            for full_name, version in dependencies.items():
                namespace, name = full_name.split(".")
                req = (namespace, name, version)
                new_req = RequirementsFileEntry(full_name, version=version, source=None)
                if not any([req in self.already_synced, new_req in self.collection_info]):
                    self.collection_info.append(new_req)
                    tasks.append(loop.create_task(self._fetch_collection_metadata(new_req)))
            await asyncio.gather(*tasks)

        info.pop("tags")
        for attr_name, attr_value in info.items():
            if attr_value is None or attr_name not in collection_version.__dict__:
                continue
            setattr(collection_version, attr_name, attr_value)

        artifact = metadata["artifact"]
        d_artifact = DeclarativeArtifact(
            artifact=Artifact(sha256=artifact["sha256"], size=artifact["size"]),
            url=url,
            relative_path=collection_version.relative_path,
            remote=self.remote,
            deferred_download=self.deferred_download,
        )

        extra_data = {}
        if api_version != 2:  # V2 never implemented the docs-blob requests
            extra_data["docs_blob_url"] = f"{collection_version_url}docs-blob/"

        d_content = DeclarativeContent(
            content=collection_version,
            d_artifacts=[d_artifact],
            extra_data=extra_data,
        )
        await self.parsing_metadata_progress_bar.aincrement()
        await self.put(d_content)

        if signatures or marks:
            collection_version = await d_content.resolution()
            for signature in signatures:
                sig = signature["signature"]
                cv_signature = CollectionVersionSignature(
                    signed_collection=collection_version,
                    data=sig,
                    digest=hashlib.sha256(sig.encode("utf-8")).hexdigest(),
                    pubkey_fingerprint=signature["pubkey_fingerprint"],
                )
                await self.put(DeclarativeContent(content=cv_signature))

            for mark_value in marks:
                cv_mark = CollectionVersionMark(
                    marked_collection=collection_version,
                    value=mark_value,
                )
                await self.put(DeclarativeContent(content=cv_mark))

        # Process syncing CV Namespace Metadata if present
        namespace_sha = metadata["namespace"].get("metadata_sha256")
        if namespace_sha:
            namespace = collection_version.namespace

            if namespace not in self.namespaces_seen:
                if await self._add_namespace(namespace, namespace_sha):
                    self.namespaces_seen.add(namespace)

    async def _add_namespace(self, name, namespace_sha):
        """Adds A Namespace metadata content to the pipeline."""

        try:
            ns = await sync_to_async(AnsibleNamespaceMetadata.objects.get)(
                metadata_sha256=namespace_sha
            )
            await self.put(DeclarativeContent(ns))
            return True
        except AnsibleNamespaceMetadata.DoesNotExist:
            pass

        endpoint, api_version = await self._get_root_api(self.remote.url)
        namespace_url = f"{endpoint}/namespaces/{name}"
        downloader = self.remote.get_downloader(
            url=namespace_url, silence_errors_for_response_codes={404}
        )
        try:
            result = await downloader.run()
        except FileNotFoundError:
            pass
        else:
            namespace = parse_metadata(result)
            links = namespace.get("links", None)

            # clean up the galaxy API for pulp
            if links:
                namespace["links"] = {x["name"]: x["url"] for x in links}
            else:
                namespace["links"] = dict()

            for key in ("pulp_href", "groups", "id", "related_fields", "users"):
                namespace.pop(key, None)

            url = namespace.pop("avatar_url", None)

            da = (
                [
                    DeclarativeFailsafeArtifact(
                        Artifact(sha256=namespace.get("avatar_sha256")),
                        url=url,
                        remote=self.remote,
                        relative_path=f"{name}-avatar",
                        deferred_download=False,
                    )
                ]
                if url
                else None
            )

            namespace = AnsibleNamespaceMetadata(**namespace)
            dc = DeclarativeContent(namespace, d_artifacts=da)
            await self.put(dc)
            return True

        return False

    async def _add_collection_version_from_git(self, url, gitref, metadata_only):
        d_content = await declarative_content_from_git_repo(
            self.remote, url, gitref, metadata_only=False
        )
        await self.put(d_content)

    def _collection_versions_list_downloader(
        self, api_version, collection_endpoint, namespace, name, page_num, page_size
    ):
        url_without_get_params = f"{collection_endpoint}{namespace}/{name}/versions/"
        if api_version == 2:
            versions_list_url = f"{url_without_get_params}?page={page_num}&page_size={page_size}"
        else:
            offset = (page_num - 1) * page_size
            versions_list_url = f"{url_without_get_params}?limit={page_size}&offset={offset}"
        return self.remote.get_downloader(url=versions_list_url)

    async def _fetch_paginated_collection_metadata(self, name, namespace, requirement, source=None):
        root = source or self.remote.url
        collection_endpoint, api_version = await self._get_paginated_collection_api(root)
        collection_url = f"{collection_endpoint}{namespace}/{name}"
        collection_metadata_downloader = self.remote.get_downloader(url=collection_url)
        collection_metadata = parse_metadata(await collection_metadata_downloader.run())
        loop = asyncio.get_event_loop()

        tasks = []
        page_num = 1
        while True:
            versions_list_downloader = self._collection_versions_list_downloader(
                api_version, collection_endpoint, namespace, name, page_num, PAGE_SIZE
            )
            collection_versions_list = parse_metadata(await versions_list_downloader.run())
            if api_version == 2:
                collection_versions = collection_versions_list["results"]
            else:
                collection_versions = collection_versions_list["data"]
            for collection_version in collection_versions:
                if Version(collection_version["version"]) in requirement:
                    version_num = collection_version["version"]
                    collection_version_detail_url = f"{collection_url}/versions/{version_num}/"
                    if collection_metadata["deprecated"]:
                        d_content = DeclarativeContent(
                            content=AnsibleCollectionDeprecated(namespace=namespace, name=name),
                        )
                        self.deprecation_after_sync.add(f"{namespace}.{name}")
                        await self.put(d_content)
                    tasks.append(
                        loop.create_task(
                            self._fetch_collection_version_metadata(
                                api_version,
                                collection_version_detail_url,
                            )
                        )
                    )
            next_value = self._get_response_next_value(api_version, collection_versions_list)
            if not next_value:
                break
            page_num = page_num + 1

        self.parsing_metadata_progress_bar.total += len(tasks)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        await asyncio.gather(*tasks)

    async def _read_from_downloaded_metadata(self, name, namespace, requirement):
        tasks = []
        loop = asyncio.get_event_loop()

        if (
            namespace not in self._unpaginated_collection_metadata
            or name not in self._unpaginated_collection_metadata[namespace]
        ):
            raise FileNotFoundError(
                f"Collection {namespace}.{name} does not exist on {self.remote.url}"
            )

        if self._unpaginated_collection_metadata[namespace][name]["deprecated"]:
            d_content = DeclarativeContent(
                content=AnsibleCollectionDeprecated(namespace=namespace, name=name),
            )
            self.deprecation_after_sync.add(f"{namespace}.{name}")
            await self.put(d_content)

        all_versions_of_collection = self._unpaginated_collection_version_metadata[namespace][name]
        for col_version_metadata in all_versions_of_collection:
            if Version(col_version_metadata["version"]) in requirement:
                if "git_url" in col_version_metadata and col_version_metadata["git_url"]:
                    tasks.append(
                        loop.create_task(
                            self._add_collection_version_from_git(
                                col_version_metadata["git_url"],
                                col_version_metadata["git_commit_sha"],
                                False,
                            )
                        )
                    )
                else:
                    collection_version_url = urljoin(
                        self.remote.url, f"{col_version_metadata['href']}"
                    )
                    tasks.append(
                        loop.create_task(
                            self._add_collection_version(
                                self._api_version, collection_version_url, col_version_metadata
                            )
                        )
                    )
        self.parsing_metadata_progress_bar.total += len(tasks)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        await asyncio.gather(*tasks)

    async def _fetch_collection_metadata(self, requirements_entry):
        requirement_version = AnsibleSpec(requirements_entry.version)

        namespace, name = requirements_entry.name.split(".")

        if self._unpaginated_collection_version_metadata and requirements_entry.source is None:
            await self._read_from_downloaded_metadata(name, namespace, requirement_version)
        else:
            await self._fetch_paginated_collection_metadata(
                name, namespace, requirement_version, requirements_entry.source
            )

    @staticmethod
    def _get_response_next_value(api_version, response):
        if api_version == 2:
            return response["next"]
        else:
            return response["links"]["next"]

    def _collection_list_downloader(self, api_version, collection_endpoint, page_num, page_size):
        if api_version == 2:
            collection_list_url = f"{collection_endpoint}?page={page_num}&page_size={page_size}"
        else:
            offset = (page_num - 1) * page_size
            collection_list_url = f"{collection_endpoint}?limit={page_size}&offset={offset}"
        return self.remote.get_downloader(url=collection_list_url)

    async def _download_unpaginated_metadata(self):
        root_endpoint, api_version = await self._get_root_api(self.remote.url)
        self._api_version = api_version
        if api_version > 2:
            loop = asyncio.get_event_loop()

            collection_endpoint = f"{root_endpoint}/collections/all/"
            excludes_endpoint = f"{root_endpoint}/excludes/"
            col_downloader = self.remote.get_downloader(
                url=collection_endpoint, silence_errors_for_response_status_codes={404}
            )
            exc_downloader = self.remote.get_downloader(
                url=excludes_endpoint, silence_errors_for_response_status_codes={404}
            )
            tasks = [loop.create_task(col_downloader.run()), loop.create_task(exc_downloader.run())]
            col_results, exc_results = await asyncio.gather(*tasks, return_exceptions=True)

            if not isinstance(exc_results, FileNotFoundError):
                excludes_response = parse_metadata(exc_results)
                if excludes_response:
                    try:
                        excludes_list = parse_collections_requirements_file(excludes_response)
                    except ValidationError:
                        pass
                    else:
                        excludes = {r.name: AnsibleSpec(r.version) for r in excludes_list}
                        self.exclude_info.update(excludes)

            if not isinstance(col_results, FileNotFoundError):
                collection_metadata_list = parse_metadata(col_results)

                self._unpaginated_collection_metadata = defaultdict(dict)
                for collection in collection_metadata_list:
                    namespace = collection["namespace"]
                    name = collection["name"]
                    self._unpaginated_collection_metadata[namespace][name] = collection

                collection_version_endpoint = f"{root_endpoint}/collection_versions/all/"
                downloader = self.remote.get_downloader(url=collection_version_endpoint)
                collection_version_metadata_list = parse_metadata(await downloader.run())

                self._unpaginated_collection_version_metadata = defaultdict(
                    lambda: defaultdict(list)
                )
                for collection_version_metadata in collection_version_metadata_list:
                    namespace = collection_version_metadata["namespace"]["name"]
                    name = collection_version_metadata["name"]
                    self._unpaginated_collection_version_metadata[namespace][name].append(
                        collection_version_metadata
                    )

    async def _find_all_collections_from_unpaginated_data(self):
        tasks = []
        loop = asyncio.get_event_loop()

        for collection_namespace_dict in self._unpaginated_collection_metadata.values():
            for collection in collection_namespace_dict.values():
                if collection["deprecated"]:
                    d_content = DeclarativeContent(
                        content=AnsibleCollectionDeprecated(
                            namespace=collection["namespace"], name=collection["name"]
                        ),
                    )
                    self.deprecation_after_sync.add(
                        f"{collection['namespace']}.{collection['name']}"
                    )
                    await self.put(d_content)

        for collections_in_namespace in self._unpaginated_collection_version_metadata.values():
            for collection_versions in collections_in_namespace.values():
                for collection_version in collection_versions:
                    collection_version_url = urljoin(
                        self.remote.url, f"{collection_version['href']}"
                    )
                    tasks.append(
                        loop.create_task(
                            self._add_collection_version(
                                self._api_version, collection_version_url, collection_version
                            )
                        )
                    )

        self.parsing_metadata_progress_bar.total = len(tasks)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        await asyncio.gather(*tasks)

    async def _find_all_collections(self):
        if self._unpaginated_collection_version_metadata:
            await self._find_all_collections_from_unpaginated_data()
            return

        collection_endpoint, api_version = await self._get_paginated_collection_api(self.remote.url)
        loop = asyncio.get_event_loop()

        tasks = []
        page_num = 1
        while True:
            collection_list_downloader = self._collection_list_downloader(
                api_version, collection_endpoint, page_num, PAGE_SIZE
            )
            collection_list = parse_metadata(await collection_list_downloader.run())

            if api_version == 2:
                collections = collection_list["results"]
            else:
                collections = collection_list["data"]

            for collection in collections:
                if api_version == 2:
                    namespace = collection["namespace"]["name"]
                else:
                    namespace = collection["namespace"]
                name = collection["name"]
                requirements_file = RequirementsFileEntry(
                    name=".".join([namespace, name]),
                    version="*",
                    source=None,
                )
                tasks.append(loop.create_task(self._fetch_collection_metadata(requirements_file)))

            next_value = self._get_response_next_value(api_version, collection_list)
            if not next_value:
                break
            page_num = page_num + 1

        self.parsing_metadata_progress_bar.total = len(tasks)
        await self.parsing_metadata_progress_bar.asave(update_fields=["total"])
        await asyncio.gather(*tasks)

    async def _should_we_sync(self):
        """Check last synced metadata time."""
        msg = _("no_change: Checking if remote changed since last sync.")
        async with ProgressReport(message=msg, code="sync.no_change") as noop:
            root, api_version = await self._get_root_api(self.remote.url)
            if api_version == 3:
                downloader = self.remote.get_downloader(
                    url=root, silence_errors_for_response_status_codes={404}
                )
                try:
                    metadata = parse_metadata(await downloader.run())
                except FileNotFoundError:
                    return True

                try:
                    self.last_synced_metadata_time = parse_datetime(metadata["published"])
                except KeyError:
                    return True

                if not self.optimize:
                    return True

                sources = set()
                if self.collection_info:
                    sources = {r.source for r in self.collection_info if r.source}
                sources.add(self.remote.url)
                if len(sources) > 1:
                    return True

                if self.last_synced_metadata_time == self.repository.last_synced_metadata_time:
                    noop.message = _(
                        "no-op: {remote} did not change since last sync - {published}".format(
                            remote=self.remote.url, published=self.last_synced_metadata_time
                        )
                    )
                    return False

        return True

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the ansible metadata.
        """
        tasks = []
        loop = asyncio.get_event_loop()

        msg = _("Parsing CollectionVersion Metadata")
        async with ProgressReport(message=msg, code="sync.parsing.metadata", total=0) as pb:
            self.parsing_metadata_progress_bar = pb
            await self._download_unpaginated_metadata()

            if self.collection_info:
                for requirement_entry in self.collection_info:
                    tasks.append(
                        loop.create_task(self._fetch_collection_metadata(requirement_entry))
                    )
            else:
                tasks.append(loop.create_task(self._find_all_collections()))
            await asyncio.gather(*tasks)
            # Ensure PR 'total' is correct before stage finishes
            pb.total = pb.done


class DeclarativeFailsafeArtifact(DeclarativeArtifact):
    """
    Special handling for downloading Namespace Avatar Artifacts.
    """

    async def download(self):
        """Allow download to fail, but not stop the sync."""
        artifact_copy = self.artifact
        try:
            return await super().download()
        except (ClientError, DigestValidationError) as e:
            # Reset DA so that future stages can properly handle it
            self.artifact = artifact_copy
            self.deferred_download = True
            name = self.extra_data.get("namespace")
            log.info(f"Failed to download namespace avatar: {name} - {e}, Skipping")
            return None


class DocsBlobDownloader(ArtifactDownloader):
    """
    Stage for downloading docs_blob.

    Args:
        max_concurrent_content (int): The maximum number of
            :class:`~pulpcore.plugin.stages.DeclarativeContent` instances to handle simultaneously.
            Default is 200.
        args: unused positional arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
        kwargs: unused keyword arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
    """

    async def _handle_content_unit(self, d_content):
        """Handle one content unit.

        Returns:
            The number of downloads
        """
        downloaded = 0
        content_already_saved = not d_content.content._state.adding
        if not content_already_saved and d_content.d_artifacts:
            remote = d_content.d_artifacts[0].remote
            docs_blob = d_content.extra_data.get("docs_blob_url")
            if docs_blob:
                downloaded = 1
                downloader = remote.get_downloader(
                    url=docs_blob, silence_errors_for_response_status_codes={404}
                )
                try:
                    download_result = await downloader.run()
                    with open(download_result.path, "r") as docs_blob_file:
                        docs_blob_json = json.load(docs_blob_file)
                        d_content.extra_data["docs_blob"] = docs_blob_json.get("docs_blob", {})
                except FileNotFoundError:
                    pass

        await self.put(d_content)
        return downloaded


class AnsibleContentSaver(ContentSaver):
    """
    A modification of ContentSaver stage that additionally saves Ansible plugin specific items.

    Saves Collection and Tag objects related to the CollectionVersion content unit.
    """

    def __init__(self, repository_version, *args, **kwargs):
        """Initialize CollectionContentSaver."""
        super().__init__(*args, **kwargs)
        self.repository_version = repository_version

    def _pre_save(self, batch):
        """
        Save a batch of Collection objects.

        Args:
            batch (list of :class:`~pulpcore.plugin.stages.DeclarativeContent`): The batch of
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be saved.

        """
        for d_content in batch:
            if d_content is None:
                continue
            if isinstance(d_content.content, CollectionVersion):
                info = d_content.content.natural_key_dict()
                collection, created = Collection.objects.get_or_create(
                    namespace=info["namespace"], name=info["name"]
                )

                d_content.content.collection = collection
            elif isinstance(d_content.content, AnsibleNamespaceMetadata):
                name = d_content.content.name
                namespace, created = AnsibleNamespace.objects.get_or_create(name=name)
                d_content.content.namespace = namespace
                if d_content.d_artifacts:
                    da = d_content.d_artifacts[0]
                    # Check to see if avatar failed to download, update metadata if so,
                    # so that the avatar should be attemtped to be downloaded again.
                    if da.deferred_download:
                        d_content.d_artifacts = []
                        d_content.content.avatar_sha256 = None
                        d_content.content.metadata_sha256 = None

    def _post_save(self, batch):
        """
        Save a batch of CollectionVersion, Tag objects.

        Args:
            batch (list of :class:`~pulpcore.plugin.stages.DeclarativeContent`): The batch of
                :class:`~pulpcore.plugin.stages.DeclarativeContent` objects to be saved.

        """
        for d_content in batch:
            if d_content is None:
                continue
            if not isinstance(d_content.content, CollectionVersion):
                continue
            collection_version = d_content.content
            docs_blob = d_content.extra_data.get("docs_blob", {})
            if docs_blob:
                collection_version.docs_blob = docs_blob

            for d_artifact in d_content.d_artifacts:
                artifact = d_artifact.artifact
                with artifact.file.open() as artifact_file, tarfile.open(
                    fileobj=artifact_file, mode="r"
                ) as tar:
                    runtime_metadata = get_file_obj_from_tarball(
                        tar, "meta/runtime.yml", artifact.file.name, raise_exc=False
                    )
                    if runtime_metadata:
                        runtime_yaml = yaml.safe_load(runtime_metadata)
                        if runtime_yaml:
                            collection_version.requires_ansible = runtime_yaml.get(
                                "requires_ansible"
                            )
                    manifest_data = json.load(
                        get_file_obj_from_tarball(tar, "MANIFEST.json", artifact.file.name)
                    )
                    files_data = json.load(
                        get_file_obj_from_tarball(tar, "FILES.json", artifact.file.name)
                    )
                    collection_version.manifest = manifest_data
                    collection_version.files = files_data
                    info = manifest_data["collection_info"]

                # Create the tags
                tags = info.pop("tags")
                for name in tags:
                    tag, created = Tag.objects.get_or_create(name=name)
                    collection_version.tags.add(tag)

                # Remove fields not used by this model
                info.pop("license_file")
                info.pop("readme")

                # Update with the additional data from the Collection
                for attr_name, attr_value in info.items():
                    if attr_value is None:
                        continue
                    setattr(collection_version, attr_name, attr_value)

                collection_version.is_highest = False
                collection_version.save()
                _update_highest_version(collection_version)



from gettext import gettext as _

from django.contrib.postgres.search import SearchQuery
from django.db.models import fields as db_fields
from django.db.models.expressions import F, Func
from django_filters import filters
from django.http import HttpResponseRedirect, HttpResponseNotFound
from drf_spectacular.utils import extend_schema
from rest_framework import mixins, serializers, viewsets
from rest_framework.decorators import action
from rest_framework.parsers import FormParser, MultiPartParser
from rest_framework.serializers import ValidationError as DRFValidationError

from pulpcore.plugin.actions import ModifyRepositoryActionMixin
from pulpcore.plugin.exceptions import DigestValidationError
from pulpcore.plugin.models import PulpTemporaryFile, RepositoryVersion
from pulpcore.plugin.serializers import AsyncOperationResponseSerializer
from pulpcore.plugin.tasking import dispatch
from pulpcore.plugin.viewsets import (
    DistributionViewSet,
    BaseFilterSet,
    ContentFilter,
    RemoteFilter,
    ContentViewSet,
    HyperlinkRelatedFilter,
    NamedModelViewSet,
    NAME_FILTER_OPTIONS,
    NoArtifactContentUploadViewSet,
    OperationPostponedResponse,
    ReadOnlyContentViewSet,
    RemoteViewSet,
    RepositoryViewSet,
    RepositoryVersionViewSet,
    RolesMixin,
    SingleArtifactContentUploadViewSet,
)
from pulpcore.plugin.util import extract_pk, raise_for_unknown_content_units, get_artifact_url
from pulp_ansible.app.galaxy.mixins import UploadGalaxyCollectionMixin
from pulp_ansible.app.utils import get_queryset_annotated_with_last_sync_task
from .models import (
    AnsibleCollectionDeprecated,
    AnsibleDistribution,
    AnsibleNamespaceMetadata,
    CollectionVersionMark,
    GitRemote,
    RoleRemote,
    AnsibleRepository,
    Collection,
    CollectionVersion,
    CollectionVersionSignature,
    CollectionRemote,
    Role,
    Tag,
)
from .serializers import (
    AnsibleDistributionSerializer,
    AnsibleNamespaceMetadataSerializer,
    CollectionVersionMarkSerializer,
    GitRemoteSerializer,
    RoleRemoteSerializer,
    AnsibleRepositorySerializer,
    AnsibleRepositoryMarkSerializer,
    AnsibleRepositorySyncURLSerializer,
    AnsibleRepositoryRebuildSerializer,
    AnsibleRepositorySignatureSerializer,
    CollectionSerializer,
    CollectionVersionSerializer,
    CollectionVersionSignatureSerializer,
    CollectionRemoteSerializer,
    CollectionOneShotSerializer,
    CopySerializer,
    RoleSerializer,
    TagSerializer,
    CollectionVersionCopyMoveSerializer,
)
from .tasks.collections import sync as collection_sync
from .tasks.collections import rebuild_repository_collection_versions_metadata
from .tasks.copy import copy_content, copy_or_move_and_sign
from .tasks.roles import synchronize as role_sync
from .tasks.git import synchronize as git_sync
from .tasks.mark import mark, unmark
from .tasks.signature import sign


class RoleFilter(ContentFilter):
    """
    FilterSet for Roles.
    """

    class Meta:
        model = Role
        fields = ["name", "namespace", "version"]


class RoleViewSet(ContentViewSet):
    """
    ViewSet for Role.
    """

    endpoint_name = "roles"
    queryset = Role.objects.all()
    serializer_class = RoleSerializer
    filterset_class = RoleFilter

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": ["create"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:ansible.modify_ansible_repo_content",
                    "has_required_repo_perms_on_upload:ansible.view_ansiblerepository",
                    "has_upload_param_model_or_obj_perms:core.change_upload",
                ],
            },
        ],
    }


class CollectionFilter(BaseFilterSet):
    """
    FilterSet for Ansible Collections.
    """

    namespace = filters.CharFilter(field_name="namespace")
    name = filters.CharFilter(field_name="name")

    class Meta:
        model = Collection
        fields = ["namespace", "name"]


class CollectionViewset(
    NamedModelViewSet, mixins.RetrieveModelMixin, mixins.ListModelMixin, RolesMixin
):
    """
    Viewset for Ansible Collections.
    """

    endpoint_name = "ansible/collections"
    queryset = Collection.objects.all()
    serializer_class = CollectionSerializer
    filterset_class = CollectionFilter


class CollectionVersionFilter(ContentFilter):
    """
    FilterSet for Ansible CollectionVersions.
    """

    namespace = filters.CharFilter(field_name="namespace")
    name = filters.CharFilter(field_name="name")
    is_highest = filters.BooleanFilter(field_name="is_highest")
    q = filters.CharFilter(field_name="q", method="filter_by_q")
    tags = filters.CharFilter(
        field_name="tags",
        method="filter_by_tags",
        help_text=_("Filter by comma separate list of tags that must all be matched"),
    )

    def filter_by_q(self, queryset, name, value):
        """
        Full text search provided by the 'q' option.

        Args:
            queryset: The query to add the additional full-text search filtering onto
            name: The name of the option specified, i.e. 'q'
            value: The string to search on

        Returns:
            The Django queryset that was passed in, additionally filtered by full-text search.

        """
        search_query = SearchQuery(value)
        qs = queryset.filter(search_vector=search_query)
        ts_rank_fn = Func(
            F("search_vector"),
            search_query,
            32,  # RANK_NORMALIZATION = 32
            function="ts_rank",
            output_field=db_fields.FloatField(),
        )
        return qs.annotate(rank=ts_rank_fn).order_by("-rank")

    def filter_by_tags(self, qs, name, value):
        """
        Filter queryset qs by list of tags.

        Args:
            qs (django.db.models.query.QuerySet): CollectionVersion queryset
            value (string): A comma separated list of tags

        Returns:
            Queryset of CollectionVersion that matches all tags

        """
        for tag in value.split(","):
            qs = qs.filter(tags__name=tag)
        return qs

    class Meta:
        model = CollectionVersion
        fields = ["namespace", "name", "version", "q", "is_highest", "tags"]


class CollectionRemoteFilter(RemoteFilter):
    class Meta:
        model = CollectionRemote
        fields = {"url": ["exact", "in", "icontains", "contains"], **RemoteFilter.Meta.fields}


class CollectionVersionViewSet(UploadGalaxyCollectionMixin, SingleArtifactContentUploadViewSet):
    """
    ViewSet for Ansible Collection.
    """

    endpoint_name = "collection_versions"
    queryset = CollectionVersion.objects.prefetch_related("_artifacts")
    serializer_class = CollectionVersionSerializer
    filterset_class = CollectionVersionFilter
    ordering_fields = ("pulp_created", "name", "version", "namespace")

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:ansible.modify_ansible_repo_content",
                    "has_required_repo_perms_on_upload:ansible.view_ansiblerepository",
                    "has_upload_param_model_or_obj_perms:core.change_upload",
                ],
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class SignatureFilter(ContentFilter):
    """
    A filter for signatures.
    """

    signed_collection = HyperlinkRelatedFilter(
        field_name="signed_collection",
        help_text=_("Filter signatures for collection version"),
    )
    signing_service = HyperlinkRelatedFilter(
        field_name="signing_service",
        help_text=_("Filter signatures produced by signature service"),
    )

    class Meta:
        model = CollectionVersionSignature
        fields = {
            "signed_collection": ["exact"],
            "pubkey_fingerprint": ["exact", "in"],
            "signing_service": ["exact"],
        }


class CollectionVersionSignatureViewSet(NoArtifactContentUploadViewSet):
    """
    ViewSet for looking at signature objects for CollectionVersion content.
    """

    endpoint_name = "collection_signatures"
    filterset_class = SignatureFilter
    queryset = CollectionVersionSignature.objects.all()
    serializer_class = CollectionVersionSignatureSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:ansible.modify_ansible_repo_content",
                    "has_required_repo_perms_on_upload:ansible.view_ansiblerepository",
                ],
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class CollectionVersionMarkFilter(ContentFilter):
    """
    A filter for marks.
    """

    marked_collection = HyperlinkRelatedFilter(
        field_name="marked_collection",
        help_text=_("Filter marks for collection version"),
    )
    value = filters.CharFilter(field_name="value", help_text=_("Filter marks by value"))

    class Meta:
        model = CollectionVersionMark
        fields = {
            "marked_collection": ["exact"],
            "value": ["exact", "in"],
        }


class CollectionVersionMarkViewSet(ContentViewSet):
    """
    ViewSet for looking at mark objects for CollectionVersion content.
    """

    endpoint_name = "collection_marks"
    filterset_class = CollectionVersionMarkFilter
    queryset = CollectionVersionMark.objects.all()
    serializer_class = CollectionVersionMarkSerializer


class CollectionDeprecatedViewSet(ContentViewSet):
    """
    ViewSet for AnsibleCollectionDeprecated.
    """

    endpoint_name = "collection_deprecations"
    queryset = AnsibleCollectionDeprecated.objects.all()
    serializer_class = CollectionSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_required_repo_perms_on_upload:ansible.modify_ansible_repo_content",
                    "has_required_repo_perms_on_upload:ansible.view_ansiblerepository",
                ],
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }


class AnsibleNamespaceFilter(ContentFilter):
    """
    A filter for namespaces.
    """

    class Meta:
        model = AnsibleNamespaceMetadata
        fields = {
            "name": NAME_FILTER_OPTIONS,
            "company": NAME_FILTER_OPTIONS,
            "metadata_sha256": ["exact", "in"],
        }


class AnsibleNamespaceViewSet(ReadOnlyContentViewSet):
    """
    ViewSet for AnsibleNamespace.
    """

    endpoint_name = "namespaces"
    queryset = AnsibleNamespaceMetadata.objects.all()
    serializer_class = AnsibleNamespaceMetadataSerializer
    filterset_class = AnsibleNamespaceFilter

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": "*",
                "principal": "admin",
                "effect": "allow",
            },
            {
                "action": "avatar",
                "principal": "*",
                "effect": "allow",
            },
        ],
        "creation_hooks": [],
    }

    @extend_schema(
        description="Get the logo for the this namespace.",
        responses={302: HttpResponseRedirect},
    )
    @action(detail=True, methods=["get"], serializer_class=None)
    def avatar(self, request, pk):
        """
        Tries to find a redirect link to the Namespace's avatar
        """
        ns = self.get_object()
        if artifact := ns.avatar_artifact:
            return HttpResponseRedirect(get_artifact_url(artifact))

        return HttpResponseNotFound()


class RoleRemoteViewSet(RemoteViewSet, RolesMixin):
    """
    ViewSet for Role Remotes.
    """

    endpoint_name = "role"
    queryset = RoleRemote.objects.all()
    serializer_class = RoleRemoteSerializer
    queryset_filtering_required_permission = "ansible.view_roleremote"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.add_roleremote",
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.delete_roleremote",
            },
            {
                "action": ["update", "partial_update", "set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.change_roleremote",
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.manage_roles_roleremote",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "ansible.roleremote_owner"},
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "ansible.roleremote_creator": ["ansible.add_roleremote"],
        "ansible.roleremote_owner": [
            "ansible.view_roleremote",
            "ansible.change_roleremote",
            "ansible.delete_roleremote",
            "ansible.manage_roles_roleremote",
        ],
        "ansible.roleremote_viewer": ["ansible.view_roleremote"],
    }


class GitRemoteViewSet(RemoteViewSet, RolesMixin):
    """
    ViewSet for Ansible Remotes.

    This is a tech preview feature. The functionality may change in the future.
    """

    endpoint_name = "git"
    queryset = GitRemote.objects.all()
    serializer_class = GitRemoteSerializer
    queryset_filtering_required_permission = "ansible.view_gitremote"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.add_gitremote",
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.delete_gitremote",
            },
            {
                "action": ["update", "partial_update", "set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.change_gitremote",
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.manage_roles_gitremote",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "ansible.gitremote_owner"},
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "ansible.gitremote_creator": ["ansible.add_gitremote"],
        "ansible.gitremote_owner": [
            "ansible.view_gitremote",
            "ansible.change_gitremote",
            "ansible.delete_gitremote",
            "ansible.manage_roles_gitremote",
        ],
        "ansible.gitremote_viewer": ["ansible.view_gitremote"],
    }


class AnsibleRepositoryViewSet(RepositoryViewSet, ModifyRepositoryActionMixin, RolesMixin):
    """
    ViewSet for Ansible Repositories.
    """

    endpoint_name = "ansible"
    queryset = AnsibleRepository.objects.all()
    serializer_class = AnsibleRepositorySerializer

    def get_queryset(self):
        qs = super().get_queryset()
        action = getattr(self, "action", "")
        if action == "list" or action == "retrieve":
            qs = get_queryset_annotated_with_last_sync_task(qs)

        return qs

    queryset_filtering_required_permission = "ansible.view_ansiblerepository"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.add_ansiblerepository",
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.delete_ansiblerepository",
            },
            {
                "action": ["update", "partial_update", "set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.change_ansiblerepository",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["modify", "mark", "unmark"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.modify_ansible_repo_content",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["rebuild_metadata"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.rebuild_metadata_ansiblerepository",
                ],
            },
            {
                "action": ["repair"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.repair_ansiblerepository"
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["sign"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.sign_ansiblerepository",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["sync"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.sync_ansiblerepository",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                    # TODO Update has_remote_param_model_or_obj to handle multiple remote types
                    # "has_remote_param_model_or_obj_perms:ansible.view_collectionremote",
                    # "has_remote_param_model_or_obj_perms:ansible.view_gitremote",
                    # "has_remote_param_model_or_obj_perms:ansible.view_roleremote",
                ],
            },
            {
                "action": ["copy_collection_version"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:modify_ansible_repo_content",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
                # TODO: create a custom access condition to ensure user has perms on dest repos
            },
            {
                "action": ["move_collection_version"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:modify_ansible_repo_content",
                    "has_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
                # TODO: create a custom access condition to ensure user has perms on dest repos
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.manage_roles_ansiblerepository",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "ansible.ansiblerepository_owner"},
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "ansible.ansiblerepository_creator": ["ansible.add_ansiblerepository"],
        "ansible.ansiblerepository_owner": [
            "ansible.view_ansiblerepository",
            "ansible.change_ansiblerepository",
            "ansible.delete_ansiblerepository",
            "ansible.manage_roles_ansiblerepository",
            "ansible.modify_ansible_repo_content",
            "ansible.rebuild_metadata_ansiblerepository",
            "ansible.repair_ansiblerepository",
            "ansible.sign_ansiblerepository",
            "ansible.sync_ansiblerepository",
        ],
        "ansible.ansiblerepository_viewer": ["ansible.view_ansiblerepository"],
    }

    @extend_schema(
        description="Trigger an asynchronous task to sync Ansible content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositorySyncURLSerializer)
    def sync(self, request, pk):
        """
        Dispatches a sync task.
        """
        repository = self.get_object()
        serializer = AnsibleRepositorySyncURLSerializer(
            data=request.data, context={"request": request, "repository_pk": repository.pk}
        )
        serializer.is_valid(raise_exception=True)

        remote = serializer.validated_data.get("remote", repository.remote)
        remote = remote.cast()

        mirror = serializer.validated_data["mirror"]
        sync_kwargs = {
            "remote_pk": remote.pk,
            "repository_pk": repository.pk,
            "mirror": mirror,
        }

        if isinstance(remote, RoleRemote):
            sync_func = role_sync
        elif isinstance(remote, CollectionRemote):
            sync_func = collection_sync
            sync_kwargs["optimize"] = serializer.validated_data["optimize"]
        elif isinstance(remote, GitRemote):
            sync_func = git_sync

        result = dispatch(
            sync_func,
            exclusive_resources=[repository],
            shared_resources=[remote],
            kwargs=sync_kwargs,
        )
        return OperationPostponedResponse(result, request)

    @extend_schema(
        description="Trigger an asynchronous task to rebuild Ansible content meta.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositoryRebuildSerializer)
    def rebuild_metadata(self, request, pk):
        """
        Dispatches a collection version rebuild task.
        """
        repository = self.get_object()
        serializer = AnsibleRepositoryRebuildSerializer(
            data=request.data, context={"request": request, "repository_pk": repository.pk}
        )
        serializer.is_valid(raise_exception=True)

        result = dispatch(
            rebuild_repository_collection_versions_metadata,
            exclusive_resources=[],
            shared_resources=[repository],
            args=[repository.latest_version().pk],
            kwargs=serializer.data,
        )
        return OperationPostponedResponse(result, request)

    @extend_schema(
        description="Trigger an asynchronous task to sign Ansible content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositorySignatureSerializer)
    def sign(self, request, pk):
        """
        Dispatches a sign task.

        This endpoint is in tech preview and can change at any time in the future.
        """
        content_units = {}

        repository = self.get_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        signing_service = serializer.validated_data["signing_service"]
        content = serializer.validated_data["content_units"]

        if "*" in content:
            content_units_pks = ["*"]
        else:
            for url in content:
                content_units[extract_pk(url)] = url
            content_units_pks = list(content_units.keys())
            existing_content_units = CollectionVersion.objects.filter(pk__in=content_units_pks)
            raise_for_unknown_content_units(existing_content_units, content_units)

        result = dispatch(
            sign,
            exclusive_resources=[repository],
            kwargs={
                "repository_href": repository.pk,
                "content_hrefs": content_units_pks,
                "signing_service_href": signing_service.pk,
            },
        )
        return OperationPostponedResponse(result, request)

    def _handle_copy_or_move(self, request, copy_or_move):
        repository = self.get_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        dest_repos_pks = [x.pk for x in data["destination_repositories"]]
        reserved = data["destination_repositories"]
        shared = [repository]
        signing_service = data.get("signing_service", None)
        if signing_service:
            signing_service = signing_service.pk

        result = dispatch(
            copy_or_move_and_sign,
            exclusive_resources=reserved,
            shared_resources=shared,
            kwargs={
                "src_repo_pk": repository.pk,
                "cv_pk_list": [x.pk for x in data["collection_versions"]],
                "dest_repo_list": dest_repos_pks,
                "copy_or_move": copy_or_move,
                "signing_service_pk": signing_service,
            },
        )
        return OperationPostponedResponse(result, request)

    @extend_schema(
        description="Trigger an asynchronous task to copy collection versions.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=CollectionVersionCopyMoveSerializer)
    def copy_collection_version(self, request, pk):
        """
        Copy a list of collection versions and all of their associated content from this
        repository.
        """

        return self._handle_copy_or_move(request, "copy")

    @extend_schema(
        description="Trigger an asynchronous task to move collection versions.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=CollectionVersionCopyMoveSerializer)
    def move_collection_version(self, request, pk):
        """
        Move a list of collection versions and all of their associated content from this
        repository.
        """

        return self._handle_copy_or_move(request, "move")

    def _handle_mark_task(self, request, task_function):
        """
        Dispatches the `task_function` passed either by `mark` or `unmark` actions.
        """
        content_units = {}

        repository = self.get_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        value = serializer.validated_data["value"]
        content = serializer.validated_data["content_units"]

        if "*" in content:
            content_units_pks = ["*"]
        else:
            for url in content:
                content_units[extract_pk(url)] = url
            content_units_pks = list(content_units.keys())
            existing_content_units = CollectionVersion.objects.filter(pk__in=content_units_pks)
            raise_for_unknown_content_units(existing_content_units, content_units)

        result = dispatch(
            task_function,  # will be `mark` or `unmark`
            exclusive_resources=[repository],
            kwargs={
                "repository_href": repository.pk,
                "content_hrefs": content_units_pks,
                "value": value,
            },
        )
        return OperationPostponedResponse(result, request)

    @extend_schema(
        description="Trigger an asynchronous task to mark Ansible content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositoryMarkSerializer)
    def mark(self, request, pk):
        """
        Dispatches a mark task.
        """
        return self._handle_mark_task(request, mark)

    @extend_schema(
        description="Trigger an asynchronous task to unmark Ansible content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositoryMarkSerializer)
    def unmark(self, request, pk):
        """
        Dispatches the unmark task.

        Removes a mark from the repository, value and content_units are required
        then only marks having `marked_collection` in the content_units list
        will be removed from the repo.

        The CollectionVersionMark object will not be deleted, because
        it may be present in another repo, so we just remove from this repo
        and let orphan cleanup take care of deletion when it is eventually
        needed.
        """
        return self._handle_mark_task(request, unmark)


class AnsibleRepositoryVersionViewSet(RepositoryVersionViewSet):
    """
    AnsibleRepositoryVersion represents a single file repository version.
    """

    parent_viewset = AnsibleRepositoryViewSet

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_repository_model_or_obj_perms:ansible.view_ansiblerepository",
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_repository_model_or_obj_perms:ansible.delete_ansiblerepository",
                    "has_repository_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["repair"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_repository_model_or_obj_perms:ansible.repair_ansiblerepository",
                    "has_repository_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": ["rebuild_metadata"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_repository_model_or_obj_perms:ansible.rebuild_metadata_ansiblerepository",
                ],
            },
        ],
    }

    @extend_schema(
        description="Trigger an asynchronous task to rebuild Ansible content meta.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @action(detail=True, methods=["post"], serializer_class=AnsibleRepositoryRebuildSerializer)
    def rebuild_metadata(self, request, *args, **kwargs):
        """
        Dispatches a collection version rebuild task.
        """
        repository_version = self.get_object()
        serializer = AnsibleRepositoryRebuildSerializer(
            data=request.data,
            context={"request": request, "repository_version_pk": repository_version.pk},
        )
        serializer.is_valid(raise_exception=True)

        result = dispatch(
            rebuild_repository_collection_versions_metadata,
            exclusive_resources=[],
            shared_resources=[repository_version.repository],
            args=[repository_version.pk],
            kwargs=serializer.data,
        )
        return OperationPostponedResponse(result, request)


class CollectionRemoteViewSet(RemoteViewSet, RolesMixin):
    """
    ViewSet for Collection Remotes.
    """

    endpoint_name = "collection"
    queryset = CollectionRemote.objects.all()
    serializer_class = CollectionRemoteSerializer
    filterset_class = CollectionRemoteFilter

    def get_queryset(self):
        qs = super().get_queryset()
        action = getattr(self, "action", "")
        if action == "list" or action == "retrieve":
            qs = get_queryset_annotated_with_last_sync_task(qs)

        return qs

    queryset_filtering_required_permission = "ansible.view_collectionremote"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_perms:ansible.add_collectionremote",
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.delete_collectionremote",
            },
            {
                "action": ["update", "partial_update", "set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.change_collectionremote",
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.manage_roles_collectionremote",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "ansible.collectionremote_owner"},
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "ansible.collectionremote_creator": ["ansible.add_collectionremote"],
        "ansible.collectionremote_owner": [
            "ansible.view_collectionremote",
            "ansible.change_collectionremote",
            "ansible.delete_collectionremote",
            "ansible.manage_roles_collectionremote",
        ],
        "ansible.collectionremote_viewer": ["ansible.view_collectionremote"],
    }

    def async_reserved_resources(self, instance):
        if instance is None:
            return []
        lock = [instance]
        repos = AnsibleRepository.objects.filter(
            remote_id=instance.pk, last_synced_metadata_time__isnull=False
        )
        lock.extend(repos)
        return lock


class CollectionUploadViewSet(viewsets.ViewSet, UploadGalaxyCollectionMixin):
    """
    ViewSet for One Shot Collection Upload.

    Args:
        file@: package to upload

    """

    serializer_class = CollectionOneShotSerializer
    parser_classes = (MultiPartParser, FormParser)

    @extend_schema(
        description="Create an artifact and trigger an asynchronous task to create "
        "Collection content from it.",
        summary="Upload a collection",
        operation_id="upload_collection",
        request=CollectionOneShotSerializer,
        responses={202: AsyncOperationResponseSerializer},
        deprecated=True,
    )
    def create(self, request):
        """
        Dispatch a Collection creation task.
        """
        serializer = CollectionOneShotSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)

        expected_digests = {}
        if serializer.validated_data["sha256"]:
            expected_digests["sha256"] = serializer.validated_data["sha256"]
        try:
            temp_file = PulpTemporaryFile.init_and_validate(
                serializer.validated_data["file"],
                expected_digests=expected_digests,
            )
        except DigestValidationError:
            raise serializers.ValidationError(
                _("The provided sha256 value does not match the sha256 of the uploaded file.")
            )

        temp_file.save()
        async_result = self._dispatch_import_collection_task(temp_file.pk)

        return OperationPostponedResponse(async_result, request)


class AnsibleDistributionViewSet(DistributionViewSet, RolesMixin):
    """
    ViewSet for Ansible Distributions.
    """

    endpoint_name = "ansible"
    queryset = AnsibleDistribution.objects.all()
    serializer_class = AnsibleDistributionSerializer

    queryset_filtering_required_permission = "ansible.view_ansibledistribution"

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve", "my_permissions"],
                "principal": "authenticated",
                "effect": "allow",
            },
            {
                "action": "create",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_perms:ansible.add_ansibledistribution",
                    "has_repo_or_repo_ver_param_model_or_obj_perms:ansible.view_ansiblerepository",
                ],
            },
            {
                "action": "destroy",
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.delete_ansibledistribution",
                ],
            },
            {
                "action": ["update", "partial_update", "set_label", "unset_label"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": [
                    "has_model_or_obj_perms:ansible.change_ansibledistribution",
                ],
            },
            {
                "action": ["list_roles", "add_role", "remove_role"],
                "principal": "authenticated",
                "effect": "allow",
                "condition": "has_model_or_obj_perms:ansible.manage_roles_ansibledistribution",
            },
        ],
        "creation_hooks": [
            {
                "function": "add_roles_for_object_creator",
                "parameters": {"roles": "ansible.ansibledistribution_owner"},
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }

    LOCKED_ROLES = {
        "ansible.ansibledistribution_creator": ["ansible.add_ansibledistribution"],
        "ansible.ansibledistribution_owner": [
            "ansible.view_ansibledistribution",
            "ansible.change_ansibledistribution",
            "ansible.delete_ansibledistribution",
            "ansible.manage_roles_ansibledistribution",
        ],
        "ansible.ansibledistribution_viewer": ["ansible.view_ansibledistribution"],
    }


class TagViewSet(NamedModelViewSet, mixins.ListModelMixin):
    """
    ViewSet for Tag models.
    """

    endpoint_name = "pulp_ansible/tags"
    queryset = Tag.objects.all()
    serializer_class = TagSerializer


class CopyViewSet(viewsets.ViewSet):
    """
    ViewSet for Content Copy.
    """

    serializer_class = CopySerializer

    @extend_schema(
        description="Trigger an asynchronous task to copy ansible content from one repository "
        "into another, creating a new repository version.",
        summary="Copy content",
        operation_id="copy_content",
        request=CopySerializer,
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request):
        """Copy content."""
        serializer = CopySerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)

        config = serializer.validated_data["config"]

        config, exclusive_resources, shared_resources = self._process_config(config)

        async_result = dispatch(
            copy_content,
            exclusive_resources=exclusive_resources,
            shared_resources=shared_resources,
            args=[config],
        )
        return OperationPostponedResponse(async_result, request)

    def _process_config(self, config):
        """
        Change the hrefs into pks within config.

        This method also implicitly validates that the hrefs map to objects and it returns a list of
        repos so that the task can lock on them.
        """
        result = []
        exclusive_resources = []
        shared_resources = []

        for entry in config:
            r = dict()
            source_version = NamedModelViewSet().get_resource(
                entry["source_repo_version"], RepositoryVersion
            )
            dest_repo = NamedModelViewSet().get_resource(entry["dest_repo"], AnsibleRepository)
            r["source_repo_version"] = source_version.pk
            r["dest_repo"] = dest_repo.pk
            exclusive_resources.append(dest_repo)
            shared_resources.append(source_version.repository)

            if "dest_base_version" in entry:
                try:
                    r["dest_base_version"] = dest_repo.versions.get(
                        number=entry["dest_base_version"]
                    ).pk
                except RepositoryVersion.DoesNotExist:
                    message = _(
                        "Version {version} does not exist for repository " "'{repo}'."
                    ).format(version=entry["dest_base_version"], repo=dest_repo.name)
                    raise DRFValidationError(detail=message)

            if entry.get("content") is not None:
                r["content"] = []
                for c in entry["content"]:
                    r["content"].append(extract_pk(c))
            result.append(r)

        return result, exclusive_resources, shared_resources



"""Tests related to upload of collections."""

import hashlib
from tempfile import NamedTemporaryFile

from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, PulpTestCase
from pulp_smash.utils import http_get

from pulpcore.client.pulp_ansible import (
    AnsibleCollectionsApi,
    ContentCollectionVersionsApi,
)

from pulp_ansible.tests.functional.utils import gen_ansible_client
from pulp_ansible.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


# TODO: Check if this url will stay around
COLLECTION_URL = "https://galaxy.ansible.com/download/pulp-squeezer-0.0.9.tar.gz"


class UploadCollectionTestCase(PulpTestCase):
    """Upload a collection."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        delete_orphans()
        cls.client = gen_ansible_client()
        cls.collections_api = AnsibleCollectionsApi(cls.client)
        cls.content_api = ContentCollectionVersionsApi(cls.client)

    @classmethod
    def tearDownClass(cls):
        """Clean up after tests."""
        delete_orphans()

    def upload_collection(self, url=COLLECTION_URL):
        """Uploada collection."""
        collection_content = http_get(url)
        self.collection_sha256 = hashlib.sha256(collection_content).hexdigest()
        with NamedTemporaryFile() as temp_file:
            temp_file.write(collection_content)
            response = self.collections_api.upload_collection(file=temp_file.name)
            collection_version_href = monitor_task(response.task).created_resources[0]
            return self.content_api.read(collection_version_href)

    def test_collection_upload(self):
        """Upload a collection.

        This test targets the following issue:

        * `Pulp #5262 <https://pulp.plan.io/issues/5262>`_
        """
        response = self.upload_collection()
        self.assertEqual(response.sha256, self.collection_sha256, response)

        delete_orphans()



"""Tests that CRUD remotes."""

import pytest
from pulpcore.client.pulp_ansible import ApiException


@pytest.mark.parallel
def test_crud_git_remote(ansible_remote_git_api_client, ansible_git_remote_factory, monitor_task):
    """Create, Read, Partial Update, and Delete an Ansible Git remote."""
    # Create Git remote
    body = {"url": "https://github.com/geerlingguy/ansible-role-adminer.git"}
    remote = ansible_git_remote_factory(**body)

    # Read Git remote
    remote = ansible_remote_git_api_client.read(remote.pulp_href)
    for k, v in body.items():
        assert body[k] == getattr(remote, k)

    # Update Git remote
    update_body = {"url": "https://github.com/geerlingguy/ansible-role-ansible.git"}
    remote_update = ansible_remote_git_api_client.partial_update(remote.pulp_href, update_body)
    monitor_task(remote_update.task)
    remote = ansible_remote_git_api_client.read(remote.pulp_href)
    assert remote.url == update_body["url"]
    assert remote.metadata_only is False

    # Ensure Git remote doesn't have a download policy
    assert not hasattr(remote, "policy")

    # Delete Git remote
    remote_delete = ansible_remote_git_api_client.delete(remote.pulp_href)
    monitor_task(remote_delete.task)
    with pytest.raises(ApiException):
        ansible_remote_git_api_client.read(remote.pulp_href)


@pytest.mark.parallel
def test_git_metadata_only_remote(ansible_git_remote_factory):
    """Create a remote where `metadata_only` is set to True."""
    body = {
        "url": "https://github.com/geerlingguy/ansible-role-adminer.git",
        "metadata_only": True,
    }
    remote = ansible_git_remote_factory(**body)
    for k, v in body.items():
        assert body[k] == getattr(remote, k)



# Generated by Django 2.2.24 on 2021-07-07 17:46

from collections import defaultdict
from django.db import migrations, models
import django.db.models.deletion


def migrate_data_from_old_model_to_new_model_up(apps, schema_editor):
    """ Move objects from AnsibleCollectionDeprecated to NewAnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    NewAnsibleCollectionDeprecated = apps.get_model('ansible', 'NewAnsibleCollectionDeprecated')
    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')
    latest_repo_versions = []
    for repo in AnsibleRepository.objects.all():
        latest_repo_versions.append(repo.versions.exclude(complete=False).latest().pk)

    for deprecated_collection in AnsibleCollectionDeprecated.objects.filter(repository_version_id__in=latest_repo_versions):
        NewAnsibleCollectionDeprecated(
            namespace=deprecated_collection.collection.namespace,
            name=deprecated_collection.collection.name,
            pulp_type="ansible.collection_deprecation",
            collection_id=deprecated_collection.collection_id,
            repository_id=deprecated_collection.repository_version.repository.pk,
            version_added_id=deprecated_collection.repository_version.pk,
        ).save()


def migrate_data_from_old_model_to_new_model_down(apps, schema_editor):
    """ Move objects from NewAnsibleCollectionDeprecated to AnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    NewAnsibleCollectionDeprecated = apps.get_model('ansible', 'NewAnsibleCollectionDeprecated')
    RepositoryVersionContentDetails = apps.get_model('core', 'RepositoryVersionContentDetails')
    for deprecation in NewAnsibleCollectionDeprecated.objects.all():
        AnsibleCollectionDeprecated(
            collection_id=deprecation.collection_id,
            repository_version_id=deprecation.version_added_id,
        ).save()
        deprecation.version_memberships.all().delete()
    RepositoryVersionContentDetails.objects.filter(content_type="ansible.collection_deprecation").delete()


def create_repository_content_for_deprecation(apps, schema_editor):
    """ Create RepositoryContent for AnsibleCollectionDeprecated."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    RepositoryContent = apps.get_model('core', 'RepositoryContent')
    RepositoryVersionContentDetails = apps.get_model('core', 'RepositoryVersionContentDetails')
    repo_content = []
    counts_list = []
    count = defaultdict(int)

    # Create RepositoryContent with deduplicated deprecation items
    for name, namespace in AnsibleCollectionDeprecated.objects.values_list("name", "namespace").distinct():
        new_deprecation = AnsibleCollectionDeprecated.objects.create(pulp_type="ansible.collection_deprecation", name=name, namespace=namespace)
        for deprecation in AnsibleCollectionDeprecated.objects.filter(name=name, namespace=namespace, repository_id__isnull=False):
            repo_content.append(RepositoryContent(
                    content=new_deprecation,
                    repository_id=deprecation.repository_id,
                    version_added_id=deprecation.version_added_id,
                ))
            count[deprecation.version_added_id] += 1
            if len(repo_content) >= 1024:
                RepositoryContent.objects.bulk_create(repo_content)
                repo_content.clear()
    RepositoryContent.objects.bulk_create(repo_content)
    repo_content.clear()
    AnsibleCollectionDeprecated.objects.filter(repository_id__isnull=False).delete()

    # Update Repository content counts
    for repo_version_id, total in count.items():
        count_obj = RepositoryVersionContentDetails(
            content_type="ansible.collection_deprecation",
            repository_version_id=repo_version_id,
            count=total,
            count_type="A", # Added
        )
        counts_list.append(count_obj)

        count_obj = RepositoryVersionContentDetails(
            content_type="ansible.collection_deprecation",
            repository_version_id=repo_version_id,
            count=total,
            count_type="P", # Present
        )
        counts_list.append(count_obj)
        if len(counts_list) >= 1024:
            RepositoryVersionContentDetails.objects.bulk_create(counts_list)
            counts_list.clear()

    RepositoryVersionContentDetails.objects.bulk_create(counts_list)


def create_deprecation_from_repository_content(apps, schema_editor):
    """ Create AnsibleCollectionDeprecated from RepositoryContent."""
    AnsibleCollectionDeprecated = apps.get_model('ansible', 'AnsibleCollectionDeprecated')
    Collection = apps.get_model('ansible', 'Collection')

    for deprecation in AnsibleCollectionDeprecated.objects.all():
        col = Collection.objects.get(namespace=deprecation.namespace, name=deprecation.name)
        for repo_content in deprecation.version_memberships.all():
            AnsibleCollectionDeprecated.objects.create(
                name=deprecation.name,
                namespace=deprecation.namespace,
                collection_id=col.pk,
                repository_id=repo_content.repository_id,
                version_added_id = repo_content.version_added_id,
            )

    AnsibleCollectionDeprecated.objects.filter(repository_id__isnull=True).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0068_add_timestamp_of_interest'),
        ('ansible', '0034_handle_jsonfield_warnings'),
    ]

    operations = [
        migrations.CreateModel(
            name='NewAnsibleCollectionDeprecated',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_ansiblecollectiondeprecated', serialize=False, to='core.Content')),
                ('namespace', models.CharField(editable=False, max_length=64)),
                ('name', models.CharField(editable=False, max_length=64)),
                ('collection_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
                ('repository_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
                ('version_added_id',  models.UUIDField(editable=False, primary_key=False, serialize=False, null=True)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.content',),
        ),
        migrations.RunPython(
            code=migrate_data_from_old_model_to_new_model_up,
            reverse_code=migrate_data_from_old_model_to_new_model_down,
        ),
        migrations.DeleteModel(
            name='AnsibleCollectionDeprecated',
        ),
        migrations.RenameModel(
            old_name='NewAnsibleCollectionDeprecated',
            new_name='AnsibleCollectionDeprecated',
        ),
        migrations.RunPython(
            code=create_repository_content_for_deprecation,
            reverse_code=create_deprecation_from_repository_content,
        ),
    ]



"""Tests related to sync ansible plugin collection content type."""

import os
import unittest

from pulpcore.client.pulp_ansible import (
    AnsibleRepositorySyncURL,
    ContentCollectionVersionsApi,
    DistributionsAnsibleApi,
    PulpAnsibleApiV3CollectionsApi,
    RepositoriesAnsibleApi,
    RemotesCollectionApi,
)
from pulp_smash.pulp3.bindings import PulpTaskError, PulpTestCase

from pulp_ansible.tests.functional.utils import (
    gen_ansible_client,
    gen_ansible_remote,
    monitor_task,
    tasks,
)
from pulp_ansible.tests.functional.utils import SyncHelpersMixin, TestCaseUsingBindings
from pulp_ansible.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class SyncCollectionsFromPulpServerTestCase(TestCaseUsingBindings, SyncHelpersMixin):
    """
    Test whether one can sync collections from a Pulp server.

    This performs two sync's, the first uses the V2 API and galaxy.ansible.com. The second is from
    Pulp using the V3 API and uses the content brought in from the first sync.

    """

    def setUp(self):
        """Set up the Sync tests."""
        self.requirements_file = "collections:\n  - testing.k8s_demo_collection"
        body = gen_ansible_remote(
            url="https://galaxy.ansible.com",
            requirements_file=self.requirements_file,
            sync_dependencies=False,
        )
        self.remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, self.remote.pulp_href)

        self.first_repo = self._create_repo_and_sync_with_remote(self.remote)
        self.distribution = self._create_distribution_from_repo(self.first_repo)

    def test_sync_collections_from_pulp(self):
        """Test sync collections from pulp server."""
        second_body = gen_ansible_remote(
            url=self.distribution.client_url,
            requirements_file=self.requirements_file,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        second_remote = self.remote_collection_api.create(second_body)
        self.addCleanup(self.remote_collection_api.delete, second_remote.pulp_href)

        second_repo = self._create_repo_and_sync_with_remote(second_remote)

        first_content = self.cv_api.list(
            repository_version=f"{self.first_repo.pulp_href}versions/1/"
        )
        self.assertGreaterEqual(len(first_content.results), 1)
        second_content = self.cv_api.list(repository_version=f"{second_repo.pulp_href}versions/1/")
        self.assertGreaterEqual(len(second_content.results), 1)

    def test_sync_collections_from_pulp_using_mirror_second_time(self):
        """Test sync collections from pulp server using a mirror option the second time."""
        body = gen_ansible_remote(
            url="https://galaxy.ansible.com",
            requirements_file="collections:\n  - testing.k8s_demo_collection",
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        first_repo = self._create_repo_and_sync_with_remote(remote)
        distribution = self._create_distribution_from_repo(first_repo)

        second_body = gen_ansible_remote(url=distribution.client_url, include_pulp_auth=True)
        second_remote = self.remote_collection_api.create(second_body)
        self.addCleanup(self.remote_collection_api.delete, second_remote.pulp_href)

        second_repo = self._create_repo_and_sync_with_remote(second_remote)

        first_content = self.cv_api.list(repository_version=f"{first_repo.pulp_href}versions/1/")
        self.assertGreaterEqual(len(first_content.results), 1)
        second_content = self.cv_api.list(repository_version=f"{second_repo.pulp_href}versions/1/")
        self.assertGreaterEqual(len(second_content.results), 1)

    def test_sync_collection_named_api(self):
        """Test sync collections from pulp server."""
        body = gen_ansible_remote(
            url="https://galaxy.ansible.com",
            requirements_file="collections:\n  - rswaf.api",
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)
        distribution = self._create_distribution_from_repo(repo)

        collection = self.collections_v3api.read("api", "rswaf", distribution.base_path)

        self.assertEqual("api", collection.name)
        self.assertEqual("rswaf", collection.namespace)

    def test_noop_resync_collections_from_pulp(self):
        """Test whether sync yields no-op when repo hasn't changed since last sync."""
        second_body = gen_ansible_remote(
            url=self.distribution.client_url,
            requirements_file=self.requirements_file,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        second_remote = self.remote_collection_api.create(second_body)
        self.addCleanup(self.remote_collection_api.delete, second_remote.pulp_href)

        second_repo = self._create_repo_with_attached_remote_and_sync(second_remote)

        second_content = self.cv_api.list(repository_version=f"{second_repo.pulp_href}versions/1/")
        self.assertGreaterEqual(len(second_content.results), 1)

        # Resync
        repository_sync_data = AnsibleRepositorySyncURL(
            remote=second_remote.pulp_href, optimize=True
        )
        sync_response = self.repo_api.sync(second_repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        second_repo = self.repo_api.read(second_repo.pulp_href)
        task = tasks.read(sync_response.task)

        msg = "no-op: {url} did not change since last sync".format(url=second_remote.url)
        messages = [r.message for r in task.progress_reports]
        self.assertIn(msg, str(messages))

    def test_noop_resync_with_mirror_from_pulp(self):
        """Test whether no-op sync with mirror=True doesn't remove repository content."""
        second_body = gen_ansible_remote(
            url=self.distribution.client_url,
            requirements_file=self.requirements_file,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        second_remote = self.remote_collection_api.create(second_body)
        self.addCleanup(self.remote_collection_api.delete, second_remote.pulp_href)

        second_repo = self._create_repo_with_attached_remote_and_sync(second_remote)

        second_content = self.cv_api.list(repository_version=f"{second_repo.pulp_href}versions/1/")
        self.assertGreaterEqual(len(second_content.results), 1)

        # Resync
        repository_sync_data = AnsibleRepositorySyncURL(
            remote=second_remote.pulp_href, optimize=True, mirror=True
        )
        sync_response = self.repo_api.sync(second_repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        second_repo = self.repo_api.read(second_repo.pulp_href)
        self.assertEqual(int(second_repo.latest_version_href[-2]), 1)
        task = tasks.read(sync_response.task)

        msg = "no-op: {url} did not change since last sync".format(url=second_remote.url)
        messages = [r.message for r in task.progress_reports]
        self.assertIn(msg, str(messages))

    def test_update_requirements_file(self):
        """Test requirements_file update."""
        body = gen_ansible_remote(
            url=self.distribution.client_url,
            requirements_file=self.requirements_file,
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_with_attached_remote_and_sync(remote)
        self.assertIsNotNone(repo.last_synced_metadata_time)

        response = self.remote_collection_api.partial_update(
            remote.pulp_href, {"requirements_file": "collections:\n  - ansible.posix"}
        )
        monitor_task(response.task)

        repo = self.repo_api.read(repo.pulp_href)
        self.assertIsNone(repo.last_synced_metadata_time)

    def test_sync_with_missing_collection(self):
        """Test that syncing with a non-present collection gives a useful error."""
        body = gen_ansible_remote(
            url=self.distribution.client_url,
            requirements_file="collections:\n  - absent.not_present",
            sync_dependencies=False,
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        with self.assertRaises(PulpTaskError) as cm:
            self._create_repo_and_sync_with_remote(remote)

        task_result = cm.exception.task.to_dict()
        msg = "absent.not_present does not exist"
        self.assertIn(msg, task_result["error"]["description"], task_result["error"]["description"])


@unittest.skipUnless(
    "AUTOMATION_HUB_TOKEN_AUTH" in os.environ,
    "'AUTOMATION_HUB_TOKEN_AUTH' env var is not defined",
)
class AutomationHubV3SyncCase(PulpTestCase, SyncHelpersMixin):
    """Test syncing from Pulp to Pulp."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_ansible_client()
        cls.repo_api = RepositoriesAnsibleApi(cls.client)
        cls.remote_collection_api = RemotesCollectionApi(cls.client)
        cls.distributions_api = DistributionsAnsibleApi(cls.client)
        cls.collections_api = PulpAnsibleApiV3CollectionsApi(cls.client)
        cls.cv_api = ContentCollectionVersionsApi(cls.client)
        cls.url = "https://cloud.redhat.com/api/automation-hub/"
        cls.aurl = (
            "https://sso.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token"
        )

    def test_sync_with_token_from_automation_hub(self):
        """Test whether we can sync with an auth token from Automation Hub."""
        body = gen_ansible_remote(
            url=self.url,
            requirements_file="collections:\n  - ansible.posix",
            auth_url=self.aurl,
            token=os.environ["AUTOMATION_HUB_TOKEN_AUTH"],
            rate_limit=10,
            tls_validation=False,
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        # Check content of both repos.
        original_content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        self.assertTrue(len(original_content.results) >= 3)  # check that we have at least 3 results

    @unittest.skip("Skipping until synclist no longer creates new repository")
    def test_syncing_with_excludes_list_from_automation_hub(self):
        """Test if syncing collections know to be on the synclist will be mirrored."""
        namespace = "autohubtest2"
        # Collections are randomly generated, in case of failure, please change the names below:
        name = "collection_dep_a_zzduuntr"
        excluded = "collection_dep_a_tworzgsx"
        excluded_version = "awcrosby.collection_test==2.1.0"
        requirements = f"""
        collections:
          - {namespace}.{name}
          - {namespace}.{excluded}
          - {excluded_version}
        """
        body = gen_ansible_remote(
            url=self.url,
            requirements_file=requirements,
            auth_url=self.aurl,
            token=os.environ["AUTOMATION_HUB_TOKEN_AUTH"],
            rate_limit=10,
            tls_validation=False,
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        # Assert that at least one CollectionVersion was downloaded
        repo_ver = f"{repo.pulp_href}versions/1/"
        content = self.cv_api.list(repository_version=repo_ver)
        self.assertTrue(content.count >= 1)

        # Assert that excluded collection was not synced
        exclude_content = self.cv_api.list(repository_version=repo_ver, name=excluded)
        self.assertTrue(exclude_content.count == 0)


# TODO QA-AH has been deprecated, remove/replace tests
@unittest.skipUnless(
    "QA_AUTOMATION_HUB_TOKEN_AUTH" in os.environ,
    "'QA_AUTOMATION_HUB_TOKEN_AUTH' env var is not defined",
)
class AutomationHubCIV3SyncCase(PulpTestCase, SyncHelpersMixin):
    """Test syncing from Pulp to Pulp."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_ansible_client()
        cls.repo_api = RepositoriesAnsibleApi(cls.client)
        cls.remote_collection_api = RemotesCollectionApi(cls.client)
        cls.distributions_api = DistributionsAnsibleApi(cls.client)
        cls.collections_api = PulpAnsibleApiV3CollectionsApi(cls.client)
        cls.cv_api = ContentCollectionVersionsApi(cls.client)
        cls.url = "https://qa.cloud.redhat.com/api/automation-hub/"
        cls.aurl = (
            "https://sso.qa.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token"
        )

    def test_mirror_from_automation_hub_ci_with_auth_token(self):
        """Test whether we can mirror from Automation Hub CI with an auth token."""
        body = gen_ansible_remote(
            url=self.url,
            auth_url=self.aurl,
            token=os.environ["QA_AUTOMATION_HUB_TOKEN_AUTH"],
            rate_limit=10,
            tls_validation=False,
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        # Assert at least one hundred CollectionVersions are returned
        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        self.assertTrue(len(content.results) >= 100)

    def test_sync_from_automation_hub_ci_with_auth_token_and_requirements_file(self):
        """Test sync from Automation Hub CI with an auth token and requirements file."""
        namespace = "autohubtest2"
        # Collections are randomly generated, in case of failure, please change the name below:
        name = "collection_dep_a_zzduuntr"
        body = gen_ansible_remote(
            url=self.url,
            requirements_file=f"collections:\n  - {namespace}.{name}",
            auth_url=self.aurl,
            token=os.environ["QA_AUTOMATION_HUB_TOKEN_AUTH"],
            rate_limit=10,
            tls_validation=False,
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        # Assert that at least one CollectionVersion was downloaded
        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        self.assertTrue(len(content.results) >= 1)

    def test_install_collection_with_invalid_token_from_automation_hub_ci(self):
        """Test whether we can mirror from Automation Hub CI with an invalid auth token."""
        body = gen_ansible_remote(
            url=self.url,
            auth_url=self.aurl,
            token="invalid token string",
            tls_validation=False,
            sync_dependencies=False,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        with self.assertRaises(PulpTaskError) as cm:
            self._create_repo_and_sync_with_remote(remote)

        task_result = cm.exception.task.to_dict()
        msg = (
            "400, message='Bad Request', url=URL("
            "'https://sso.qa.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token')"
        )
        self.assertEqual(
            msg, task_result["error"]["description"], task_result["error"]["description"]
        )



from django.contrib.postgres.search import SearchQuery
from django.db.models import fields as db_fields, Case, When, Value
from django.db.models import Q
from django.db.models.expressions import F, Func
from django_filters import (
    filters,
    FilterSet,
)
import semantic_version
from rest_framework.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from pulpcore.plugin.viewsets import LabelFilter

from pulpcore.plugin.models import RepositoryVersion
from pulp_ansible.app import models


class SemanticVersionOrderingFilter(filters.OrderingFilter):
    def filter(self, qs, value):
        if value is not None and any(v in ["version", "-version"] for v in value):
            order = "-" if "-version" in value else ""

            return qs.annotate(
                prerelease=Case(
                    When(collection_version__version_prerelease="", then=Value(None)),
                    default="collection_version__version_prerelease",
                ),
            ).order_by(
                f"{order}collection_version__version_major",
                f"{order}collection_version__version_minor",
                f"{order}collection_version__version_patch",
                f"{order}prerelease",
            )

        return super().filter(qs, value)


class CollectionVersionSearchFilter(FilterSet):
    """A custom filterset for cross-repo search."""

    strict = False

    name = filters.CharFilter(field_name="collection_version__name", lookup_expr="exact")
    namespace = filters.CharFilter(field_name="collection_version__namespace", lookup_expr="exact")
    version = filters.CharFilter(field_name="collection_version__version", lookup_expr="exact")
    repository_name = filters.CharFilter(method="filter_by_repository_name")
    repository = filters.CharFilter(method="filter_by_repository_id")
    is_highest = filters.BooleanFilter(field_name="is_highest")
    highest = filters.BooleanFilter(field_name="is_highest")
    is_deprecated = filters.BooleanFilter(field_name="is_deprecated")
    deprecated = filters.BooleanFilter(field_name="is_deprecated")
    is_signed = filters.BooleanFilter(field_name="is_signed")
    signed = filters.BooleanFilter(field_name="is_signed")
    q = filters.CharFilter(field_name="q", method="filter_by_q")
    tags = filters.CharFilter(
        field_name="tags",
        method="filter_by_tags",
        help_text="Filter by comma separate list of tags that must all be matched",
    )
    distribution = filters.CharFilter(method="filter_by_distribution_id")
    distribution_base_path = filters.CharFilter(method="filter_by_base_path")
    dependency = filters.CharFilter(method="filter_by_dependency")
    version_range = filters.CharFilter(field_name="version_range", method="version_range_filter")
    repository_version = filters.CharFilter(
        field_name="repository_version__number", method="repository_version_filter"
    )
    keywords = filters.CharFilter(field_name="q", method="filter_by_q")
    repository_label = LabelFilter(label_field_name="repository__pulp_labels")

    order_by = SemanticVersionOrderingFilter(
        choices=(
            ("pulp_created", "by CV created"),
            ("-pulp_created", "by CV created (descending)"),
            ("namespace", "by CV namespace"),
            ("-namespace", "by CV namespace (descending)"),
            ("name", "by CV name"),
            ("-name", "by CV name (descending)"),
            ("version", "by CV version"),
            ("-version", "by CV version (descending)"),
        ),
        fields={
            "collection_version__pulp_created": "pulp_created",
            "collection_version__namespace": "namespace",
            "collection_version__name": "name",
        },
    )

    def version_range_filter(self, queryset, name, value):
        try:
            s = semantic_version.SimpleSpec(value)
            full_version_list = [
                semantic_version.Version(v)
                for v in queryset.values_list("collection_version__version", flat=True)
            ]
            version_list = [str(v) for v in s.filter(full_version_list)]

            return queryset.filter(collection_version__version__in=version_list)
        except ValueError:
            raise ValidationError(_("%s must be a valid semantic version range." % name))

    def filter_by_repository_name(self, queryset, name, value):
        repository_names = self.request.query_params.getlist(name)
        return queryset.filter(repository__name__in=repository_names)

    def filter_by_repository_id(self, queryset, name, value):
        repository_ids = self.request.query_params.getlist(name)
        return queryset.filter(repository__pk__in=repository_ids)

    def filter_by_distribution_id(self, qs, name, value):
        dist_ids = self.request.query_params.getlist(name)

        query = Q()

        for dist_id in dist_ids:
            for distribution in models.AnsibleDistribution.objects.filter(pk=dist_id):
                if distribution.repository_version_id:
                    query = query | Q(repository_version__pk=distribution.repository_version_id)
                else:
                    query = query | Q(repository__pk=distribution.repository_id)

        return qs.filter(query)

    def filter_by_base_path(self, qs, name, value):
        base_paths = self.request.query_params.getlist(name)

        query = Q()

        for base_path in base_paths:
            for distribution in models.AnsibleDistribution.objects.filter(base_path=base_path):
                if distribution.repository_version_id:
                    query = query | Q(repository_version__pk=distribution.repository_version_id)
                else:
                    query = query | Q(repository__pk=distribution.repository_id)

        return qs.filter(query)

    def repository_version_filter(self, qs, name, value):
        if value != "latest":
            return qs.filter(repository_version__number=value)

        # Reduce queryset down to the "latest" repository version number(s)
        rqs = (
            RepositoryVersion.objects.filter(repository__pulp_type="ansible.ansible")
            .order_by("repository", "-pulp_created")
            .distinct("repository")
            .values_list("repository__pk", "number")
        )
        query = Q(repository_version=None)
        for rpk, number in rqs:
            query = query | Q(repository__pk=rpk, repository_version__number=number)

        return qs.filter(query)

    def filter_by_dependency(self, qs, name, value):
        """Return a list of collections that depend on a given collection name."""
        qs = qs.filter(collection_version__dependencies__has_key=value)
        return qs

    def filter_by_q(self, queryset, name, value):
        """
        Full text search provided by the 'q' option.
        Args:
            queryset: The query to add the additional full-text search filtering onto
            name: The name of the option specified, i.e. 'q'
            value: The string to search on
        Returns:
            The Django queryset that was passed in, additionally filtered by full-text search.
        """
        search_query = SearchQuery(value)
        qs = queryset.filter(collection_version__search_vector=search_query)
        ts_rank_fn = Func(
            F("collection_version__search_vector"),
            search_query,
            32,  # RANK_NORMALIZATION = 32
            function="ts_rank",
            output_field=db_fields.FloatField(),
        )
        return qs.annotate(rank=ts_rank_fn).order_by("-rank")

    def filter_by_tags(self, qs, name, value):
        """
        Filter queryset qs by list of tags.
        Args:
            qs (django.db.models.query.QuerySet): CollectionVersion queryset
            value (string): A comma separated list of tags
        Returns:
            Queryset of CollectionVersion that matches all tags
        """
        for tag in value.split(","):
            qs = qs.filter(collection_version__tags__name=tag)
        return qs



from django.conf import settings
from django.urls import include, path, re_path

from pulp_ansible.app.galaxy.views import (
    GalaxyCollectionVersionDetail,
    GalaxyCollectionVersionList,
    GalaxyCollectionDetailView,
    GalaxyCollectionView,
    GalaxyVersionView,
    RoleList,
    RoleVersionList,
)
from pulp_ansible.app.galaxy.v3 import views as views_v3
from pulp_ansible.app.galaxy.v3 import viewsets as viewsets_v3

from pulp_ansible.app.viewsets import CopyViewSet, CollectionUploadViewSet

from pulpcore.plugin.serializers import AsyncOperationResponseSerializer

GALAXY_API_ROOT = getattr(settings, "GALAXY_API_ROOT", "pulp_ansible/galaxy/<path:path>/api/")


v1_urls = [
    path("roles/", RoleList.as_view()),
    path("roles/<str:role_pk>/versions/", RoleVersionList.as_view()),
]

v2_urls = [
    path("collections/", GalaxyCollectionView.as_view()),
    path("collections/<str:namespace>/<str:name>/", GalaxyCollectionDetailView.as_view()),
    path("collections/<str:namespace>/<str:name>/versions/", GalaxyCollectionVersionList.as_view()),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/",
        GalaxyCollectionVersionDetail.as_view(),
        name="v2-collection-versions-detail",
    ),
    path(
        "collection-imports/<uuid:pk>/",
        views_v3.CollectionImportViewSet.as_view({"get": "retrieve"}),
    ),
]

# Legacy urls that need to be redirected to plugin/ansible/collections/<base_path>/
legacy_v3_collection_urls = [
    path(
        "",
        views_v3.redirect_view_generator(
            {"get": "retrieve"}, url="repo-metadata", viewset=views_v3.RepoMetadataViewSet
        ),
        name="legacy-v3-repo-metadata",
    ),
    path(
        "collections/",
        views_v3.redirect_view_generator(
            {"get": "list"}, url="collections-list", viewset=views_v3.CollectionViewSet
        ),
        name="legacy-v3-collections-list",
    ),
    path(
        "artifacts/collections/<str:path>/<str:filename>",
        views_v3.redirect_view_generator(
            {"get": "get"},
            url="collection-artifact-download",
            viewset=views_v3.CollectionArtifactDownloadView,
        ),
        name="legacy-v3-collection-artifact-download",
    ),
    path(
        "collections/<str:namespace>/<str:name>/",
        views_v3.redirect_view_generator(
            {"get": "retrieve", "patch": "update", "delete": "destroy"},
            url="collections-detail",
            viewset=views_v3.CollectionViewSet,
            responses={"update": AsyncOperationResponseSerializer},
        ),
        name="legacy-v3-collections-detail",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/",
        views_v3.redirect_view_generator(
            {"get": "list"},
            url="collection-versions-list",
            viewset=views_v3.CollectionVersionViewSet,
        ),
        name="legacy-v3-collection-versions-list",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/",
        views_v3.redirect_view_generator(
            {"get": "retrieve", "delete": "destroy"},
            url="collection-versions-detail",
            viewset=views_v3.CollectionVersionViewSet,
        ),
        name="legacy-v3-collection-versions-detail",
    ),
    path(
        "collections/<str:namespace>/<str:name>/versions/<str:version>/docs-blob/",
        views_v3.redirect_view_generator(
            {"get": "retrieve"},
            url="collection-versions-detail-docs",
            viewset=views_v3.CollectionVersionDocsViewSet,
        ),
        name="legacy-v3-collection-versions-detail-docs",
    ),
    path(
        "collections/all/",
        views_v3.redirect_view_generator(
            {"get": "list"},
            url="metadata-collection-list",
            viewset=views_v3.UnpaginatedCollectionViewSet,
        ),
        name="legacy-v3-metadata-collection-list",
    ),
    path(
        "collection_versions/all/",
        views_v3.redirect_view_generator(
            {"get": "list"},
            url="metadata-collection-versions-list",
            viewset=views_v3.UnpaginatedCollectionVersionViewSet,
        ),
        name="legacy-v3-metadata-collection-versions-list",
    ),
    path(
        "namespaces/",
        views_v3.redirect_view_generator(
            {"get": "list"},
            url="ansible-namespaces-list",
            viewset=views_v3.AnsibleNamespaceViewSet,
        ),
        name="legacy-v3-ansible-namespaces-list",
    ),
    path(
        "namespaces/<str:name>/",
        views_v3.redirect_view_generator(
            {"get": "retrieve"},
            url="ansible-namespaces-detail",
            viewset=views_v3.AnsibleNamespaceViewSet,
        ),
        name="legacy-v3-ansible-namespaces-detail",
    ),
    # the ansible-galaxy client doesn't play well with redirects for POST operations, so these
    # views don't redirect
    path(
        "artifacts/collections/",
        views_v3.LegacyCollectionUploadViewSet.as_view({"post": "create"}),
        name="collection-artifact-upload",
    ),
]


legacy_v3_urls = [
    path(
        "imports/collections/<uuid:pk>/",
        views_v3.redirect_view_generator(
            {"get": "retrieve"},
            url="collection-imports-detail",
            viewset=views_v3.CollectionImportViewSet,
            distro_view=False,
        ),
        name="legacy-v3-collection-imports-detail",
    ),
]

v3_collection_urls = [
    path("", views_v3.RepoMetadataViewSet.as_view({"get": "retrieve"}), name="repo-metadata"),
    path("index/", views_v3.CollectionViewSet.as_view({"get": "list"}), name="collections-list"),
    path(
        "artifacts/",
        views_v3.CollectionUploadViewSet.as_view({"post": "create"}),
        name="collection-artifact-upload",
    ),
    path(
        "artifacts/<str:filename>",
        views_v3.CollectionArtifactDownloadView.as_view(),
        name="collection-artifact-download",
    ),
    path(
        "all-collections/",
        views_v3.UnpaginatedCollectionViewSet.as_view({"get": "list"}),
        name="metadata-collection-list",
    ),
    path(
        "all-versions/",
        views_v3.UnpaginatedCollectionVersionViewSet.as_view({"get": "list"}),
        name="metadata-collection-versions-list",
    ),
]

v3_collection_detail_urls = [
    path(
        "",
        views_v3.CollectionViewSet.as_view(
            {"get": "retrieve", "patch": "update", "delete": "destroy"}
        ),
        name="collections-detail",
    ),
    path(
        "versions/",
        views_v3.CollectionVersionViewSet.as_view({"get": "list"}),
        name="collection-versions-list",
    ),
    path(
        "versions/<str:version>/",
        views_v3.CollectionVersionViewSet.as_view({"get": "retrieve", "delete": "destroy"}),
        name="collection-versions-detail",
    ),
    path(
        "versions/<str:version>/docs-blob/",
        views_v3.CollectionVersionDocsViewSet.as_view({"get": "retrieve"}),
        name="collection-versions-detail-docs",
    ),
]

namespace_urls = [
    path(
        "",
        views_v3.AnsibleNamespaceViewSet.as_view({"get": "list", "post": "create"}),
        name="ansible-namespaces-list",
    ),
    path(
        "<str:name>/",
        views_v3.AnsibleNamespaceViewSet.as_view(
            {"get": "retrieve", "delete": "delete", "patch": "partial_update"}
        ),
        name="ansible-namespaces-detail",
    ),
]

v3_plugin_urls = [
    # path:var captures /, so it has to have something at the end to make it work
    # correctly.
    path(
        "content/<path:distro_base_path>/collections/index/<str:namespace>/<str:name>/",
        include(v3_collection_detail_urls),
    ),
    path("content/<path:distro_base_path>/collections/", include(v3_collection_urls)),
    path("content/<path:distro_base_path>/namespaces/", include(namespace_urls)),
    path(
        "imports/collections/<uuid:pk>/",
        views_v3.CollectionImportViewSet.as_view({"get": "retrieve"}),
        name="collection-imports-detail",
    ),
    path(
        "client-configuration/",
        views_v3.ClientConfigurationView.as_view(),
        name="client-configuration-viewset",
    ),
    path(
        "search/collection-versions/",
        viewsets_v3.CollectionVersionSearchViewSet.as_view({"get": "list", "post": "rebuild"}),
        name="collection-versions-search",
    ),
]

v3_urls = [
    path("", include(legacy_v3_collection_urls)),
    path("", include(legacy_v3_urls)),
    path("plugin/ansible/", include(v3_plugin_urls)),
]

urlpatterns = [
    path("ansible/collections/", CollectionUploadViewSet.as_view({"post": "create"})),
    path(GALAXY_API_ROOT.split("<path:path>")[0] + "default/api/v3/", include(v3_urls)),
    path(
        GALAXY_API_ROOT.split("<path:path>")[0] + "default/api/",
        GalaxyVersionView.as_view(v3_only=True),
    ),
    path(GALAXY_API_ROOT + "v1/", include(v1_urls)),
    path(GALAXY_API_ROOT + "v2/", include(v2_urls)),
    path(GALAXY_API_ROOT + "v3/", include(v3_urls)),
    path(GALAXY_API_ROOT, GalaxyVersionView.as_view()),
    re_path(r"^pulp/api/v3/ansible/copy/$", CopyViewSet.as_view({"post": "create"})),
]
